import copy
import random
import warnings

import dill
import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim

from agilerl.networks.evolvable_cnn import EvolvableCNN
from agilerl.networks.evolvable_mlp import EvolvableMLP
from agilerl.wrappers.make_evolvable import MakeEvolvable


class MATD3:
    """The MATD3 algorithm class. MATD3 paper: https://arxiv.org/abs/1910.01465

    :param state_dims: State observation dimensions for each agent
    :type state_dims: list[tuple]
    :param action_dims: Action dimensions for each agent
    :type action_dims: list[int]
    :param one_hot: One-hot encoding, used with discrete observation spaces
    :type one_hot: bool
    :param n_agents: Number of agents
    :type n_agents: int
    :param agent_ids: Agent ID for each agent
    :type agent_ids: list[str]
    :param max_action: Upper bound of the action space for each agent
    :type max_action: list[float]
    :param min_action: Lower bound of the action space for each agent
    :type min_action: list[float]
    :param discrete_actions: Boolean flag to indicate a discrete action space
    :type discrete_actions: bool, optional
    :param expl_noise: Standard deviation for Gaussian exploration noise, defaults to 0.1
    :type expl_noise: float, optional
    :param policy_freq: Policy update frequency, defaults to 2
    :type policy_freq: int, optional
    :param index: Index to keep track of object instance during tournament selection and mutation, defaults to 0
    :type index: int, optional
    :param net_config: Network configuration, defaults to mlp with hidden size [64,64]
    :type net_config: dict, optional
    :param batch_size: Size of batched sample from replay buffer for learning, defaults to 64
    :type batch_size: int, optional
    :param lr: Learning rate for optimizer, defaults to 0.01
    :type lr: float, optional
    :param learn_step: Learning frequency, defaults to 5
    :type learn_step: int, optional
    :param gamma: Discount factor, defaults to 0.95
    :type gamma: float, optional
    :param tau: For soft update of target network parameters, defaults to 0.01
    :type tau: float, optional
    :param mutation: Most recent mutation to agent, defaults to None
    :type mutation: str, optional
    :param actor_networks: List of custom actor networks, defaults to None
    :type actor_networks: list[nn.Module], optional
    :param critic_networks: List containing two lists of custom critic networks, defaults to None
    :type critic_networks: list[list[nn.Module]], optional
    :param device: Device for accelerated computing, 'cpu' or 'cuda', defaults to 'cpu'
    :type device: str, optional
    :param accelerator: Accelerator for distributed computing, defaults to None
    :type accelerator: accelerate.Accelerator(), optional
    :param wrap: Wrap models for distributed training upon creation, defaults to True
    :type wrap: bool, optional
    """

    def __init__(
        self,
        state_dims,
        action_dims,
        one_hot,
        n_agents,
        agent_ids,
        max_action,
        min_action,
        discrete_actions,
        expl_noise=0.1,
        index=0,
        policy_freq=2,
        net_config={"arch": "mlp", "h_size": [64, 64]},
        batch_size=64,
        lr=0.01,
        learn_step=5,
        gamma=0.95,
        tau=0.01,
        mutation=None,
        actor_networks=None,
        critic_networks=None,
        device="cpu",
        accelerator=None,
        wrap=True,
    ):
        assert isinstance(state_dims, list), "State dimensions must be a list."
        assert isinstance(action_dims, list), "Action dimensions must be a list."
        assert isinstance(
            one_hot, bool
        ), "One-hot encoding flag must be boolean value True or False."
        assert isinstance(n_agents, int), "Number of agents must be an integer."
        assert isinstance(
            agent_ids, (tuple, list)
        ), "Agent IDs must be stores in a tuple or list."
        assert isinstance(
            discrete_actions, bool
        ), "Discrete actions flag must be a boolean value True or False."
        assert (
            isinstance(max_action, list) or max_action is None
        ), "Max action must be a list."
        assert (
            isinstance(min_action, list) or min_action is None
        ), "Min action must be a list."
        assert (max_action is not None) == (
            min_action is not None
        ), "Max and min actions must both be supplied, or both be None."
        if max_action is not None and min_action is not None:
            for x, n in zip(max_action, min_action):
                x, n = x[0], n[0]
                assert x > n, "Max action must be greater than min action."
                assert x > 0, "Max action must be greater than zero."
                assert n <= 0, "Min action must be less than or equal to zero."
        assert isinstance(index, int), "Agent index must be an integer."
        assert isinstance(batch_size, int), "Batch size must be an integer."
        assert batch_size >= 1, "Batch size must be greater than or equal to one."
        assert isinstance(lr, float), "Learning rate must be a float."
        assert lr > 0, "Learning rate must be greater than zero."
        assert isinstance(learn_step, int), "Learn step rate must be an integer."
        assert learn_step >= 1, "Learn step must be greater than or equal to one."
        assert isinstance(gamma, float), "Gamma must be a float."
        assert isinstance(tau, float), "Tau must be a float."
        assert tau > 0, "Tau must be greater than zero."
        assert isinstance(policy_freq, int), "Policy frequency must be an integer."
        assert policy_freq > 0, "Policy frequency must be grateer than zero."
        assert n_agents == len(
            agent_ids
        ), "Number of agents must be equal to the length of the agent IDs list."
        if actor_networks is not None:
            assert all(
                isinstance(actor, nn.Module) for actor in actor_networks
            ), "Actor networks must be an nn.Module or None."
        if critic_networks is not None:
            for critic_network in critic_networks:
                assert all(
                    isinstance(critic, nn.Module) for critic in critic_network
                ), "Critic networks must be an nn.Module or None."
        if (actor_networks is not None) != (critic_networks is not None):
            warnings.warn(
                "Actor and critic network lists must both be supplied to use custom networks. Defaulting to net config."
            )
        assert isinstance(
            wrap, bool
        ), "Wrap models flag must be boolean value True or False."
        self.algo = "MATD3"
        self.state_dims = state_dims
        self.total_state_dims = sum(state_dim[0] for state_dim in self.state_dims)
        self.action_dims = action_dims
        self.one_hot = one_hot
        self.n_agents = n_agents
        self.multi = True
        self.agent_ids = agent_ids
        self.net_config = net_config
        self.batch_size = batch_size
        self.lr = lr
        self.learn_step = learn_step
        self.gamma = gamma
        self.tau = tau
        self.mut = mutation
        self.device = device
        self.accelerator = accelerator
        self.index = index
        self.policy_freq = policy_freq
        self.scores = []
        self.fitness = []
        self.steps = [0]
        self.max_action = max_action
        self.expl_noise = expl_noise
        self.min_action = min_action
        self.discrete_actions = discrete_actions
        self.total_actions = sum(self.action_dims)
        self.actor_networks = actor_networks
        self.critic_networks = critic_networks

        if self.actor_networks is not None and self.critic_networks is not None:
            self.actors = actor_networks
            self.critics_1 = critic_networks[0]
            self.critics_2 = critic_networks[-1]
            self.net_config = None
        else:
            if "output_activation" in self.net_config.keys():
                pass
            else:
                if self.discrete_actions:
                    self.net_config["output_activation"] = "GumbelSoftmax"

            # model
            if self.net_config["arch"] == "mlp":  # Multi-layer Perceptron
                self.actors = []
                for idx, (action_dim, state_dim) in enumerate(
                    zip(self.action_dims, self.state_dims)
                ):
                    if not self.discrete_actions:
                        if self.min_action[idx][0] < 0:
                            self.net_config["output_activation"] = "Tanh"
                        else:
                            self.net_config["output_activation"] = "Sigmoid"
                    self.actors.append(
                        EvolvableMLP(
                            num_inputs=state_dim[0],
                            num_outputs=action_dim,
                            hidden_size=self.net_config["h_size"],
                            mlp_output_activation=self.net_config["output_activation"],
                            device=self.device,
                            accelerator=self.accelerator,
                        )
                    )
                self.critics_1 = [
                    EvolvableMLP(
                        num_inputs=self.total_state_dims + self.total_actions,
                        num_outputs=1,
                        hidden_size=self.net_config["h_size"],
                        device=self.device,
                        accelerator=self.accelerator,
                    )
                    for _ in range(self.n_agents)
                ]
                self.critics_2 = [
                    EvolvableMLP(
                        num_inputs=self.total_state_dims + self.total_actions,
                        num_outputs=1,
                        hidden_size=self.net_config["h_size"],
                        device=self.device,
                        accelerator=self.accelerator,
                    )
                    for _ in range(self.n_agents)
                ]
            elif self.net_config["arch"] == "cnn":  # Convolutional Neural Network
                self.actors = []
                for idx, (action_dim, state_dim) in enumerate(
                    zip(self.action_dims, self.state_dims)
                ):
                    if not self.discrete_actions:
                        if self.min_action[idx][0] < 0:
                            self.net_config["output_activation"] = "Tanh"
                        else:
                            self.net_config["output_activation"] = "Sigmoid"
                    self.actors.append(
                        EvolvableCNN(
                            input_shape=state_dim,
                            num_actions=action_dim,
                            channel_size=self.net_config["c_size"],
                            kernel_size=self.net_config["k_size"],
                            stride_size=self.net_config["s_size"],
                            hidden_size=self.net_config["h_size"],
                            normalize=self.net_config["normalize"],
                            mlp_output_activation=self.net_config["output_activation"],
                            multi=self.multi,
                            n_agents=self.n_agents,
                            device=self.device,
                            accelerator=self.accelerator,
                        )
                    )
                self.critics_1 = [
                    EvolvableCNN(
                        input_shape=state_dim,
                        num_actions=self.total_actions,
                        channel_size=self.net_config["c_size"],
                        kernel_size=self.net_config["k_size"],
                        stride_size=self.net_config["s_size"],
                        hidden_size=self.net_config["h_size"],
                        normalize=self.net_config["normalize"],
                        mlp_activation="Tanh",
                        mlp_output_activation=None,
                        critic=True,
                        n_agents=self.n_agents,
                        multi=self.multi,
                        device=self.device,
                        accelerator=self.accelerator,
                    )
                    for state_dim in self.state_dims
                ]
                self.critics_2 = [
                    EvolvableCNN(
                        input_shape=state_dim,
                        num_actions=self.total_actions,
                        channel_size=self.net_config["c_size"],
                        kernel_size=self.net_config["k_size"],
                        stride_size=self.net_config["s_size"],
                        hidden_size=self.net_config["h_size"],
                        normalize=self.net_config["normalize"],
                        mlp_activation="Tanh",
                        mlp_output_activation=None,
                        critic=True,
                        n_agents=self.n_agents,
                        multi=self.multi,
                        device=self.device,
                        accelerator=self.accelerator,
                    )
                    for state_dim in self.state_dims
                ]

        # Assign architecture
        self.arch = (
            self.net_config["arch"]
            if self.net_config is not None
            else self.actors[0].arch
        )

        # Create target networks
        self.actor_targets = copy.deepcopy(self.actors)
        self.critic_targets_1 = copy.deepcopy(self.critics_1)
        self.critic_targets_2 = copy.deepcopy(self.critics_2)

        # Initialise target network parameters
        for actor, actor_target in zip(self.actors, self.actor_targets):
            actor_target.load_state_dict(actor.state_dict())
        for critic_1, critic_2, critic_target_1, critic_target_2 in zip(
            self.critics_1, self.critics_2, self.critic_targets_1, self.critic_targets_2
        ):
            critic_target_1.load_state_dict(critic_1.state_dict())
            critic_target_2.load_state_dict(critic_2.state_dict())

        self.actor_optimizers_type = [
            optim.Adam(actor.parameters(), lr=self.lr) for actor in self.actors
        ]
        self.critic_1_optimizers_type = [
            optim.Adam(critic.parameters(), lr=self.lr) for critic in self.critics_1
        ]
        self.critic_2_optimizers_type = [
            optim.Adam(critic.parameters(), lr=self.lr) for critic in self.critics_2
        ]

        if self.accelerator is not None:
            self.actor_optimizers = self.actor_optimizers_type
            self.critic_1_optimizers = self.critic_1_optimizers_type
            self.critic_2_optimizers = self.critic_2_optimizers_type
            if wrap:
                self.wrap_models()
        else:
            self.actors = [actor.to(self.device) for actor in self.actors]
            self.actor_targets = [
                actor_target.to(self.device) for actor_target in self.actor_targets
            ]
            self.critics_1 = [critic.to(self.device) for critic in self.critics_1]
            self.critic_targets_1 = [
                critic_target.to(self.device) for critic_target in self.critic_targets_1
            ]
            self.critics_2 = [critic.to(self.device) for critic in self.critics_2]
            self.critic_targets_2 = [
                critic_target.to(self.device) for critic_target in self.critic_targets_2
            ]
            self.actor_optimizers = self.actor_optimizers_type
            self.critic_1_optimizers = self.critic_1_optimizers_type
            self.critic_2_optimizers = self.critic_2_optimizers_type

        self.criterion = nn.MSELoss()

    def scale_to_action_space(self, action, idx):
        """Scales actions to action space defined by self.min_action and self.max_action.

        :param action: Action to be scaled
        :type action: numpy.ndarray
        """
        return np.where(
            action > 0,
            action * self.max_action[idx][0],
            action * -self.min_action[idx][0],
        )

    def getAction(self, states, epsilon=0, agent_mask=None, env_defined_actions=None):
        """Returns the next action to take in the environment.
        Epsilon is the probability of taking a random action, used for exploration.
        For epsilon-greedy behaviour, set epsilon to 0.

        :param state: Environment observations: {'agent_0': state_dim_0, ..., 'agent_n': state_dim_n}
        :type state: Dict[str, numpy.Array]
        :param epsilon: Probablilty of taking a random action for exploration, defaults to 0
        :type epsilon: float, optional
        :param agent_mask: Mask of agents to return actions for: {'agent_0': True, ..., 'agent_n': False}
        :type agent_mask: Dict[str, bool]
        :param env_defined_actions: Dictionary of actions defined by the environment: {'agent_0': np.array, ..., 'agent_n': np.array}
        :type env_defined_actions: Dict[str, np.array]
        """
        # Get agents, states and actions we want to take actions for at this timestep according to agent_mask
        if agent_mask is None:
            agent_ids = self.agent_ids
            actors = self.actors
            state_dims = self.state_dims
        else:
            agent_ids = [agent for agent in agent_mask.keys() if agent_mask[agent]]
            state_dims = [
                state_dim
                for state_dim, mask_flag in zip(self.state_dims, agent_mask.keys())
                if mask_flag
            ]
            action_dims_dict = {
                agent: action_dim
                for agent, action_dim in zip(self.agent_ids, self.action_dims)
            }
            states = {
                agent: states[agent] for agent in agent_mask.keys() if agent_mask[agent]
            }
            actors = [
                actor
                for agent, actor in zip(agent_mask.keys(), self.actors)
                if agent_mask[agent]
            ]

        # Convert states to a list of torch tensors
        states = [torch.from_numpy(state).float() for state in states.values()]

        # Configure accelerator
        if self.accelerator is None:
            states = [state.to(self.device) for state in states]

        if self.one_hot:
            states = [
                nn.functional.one_hot(state.long(), num_classes=state_dim[0])
                .float()
                .squeeze()
                for state, state_dim in zip(states, state_dims)
            ]

        if self.arch == "mlp":
            states = [
                state.unsqueeze(0) if len(state.size()) < 2 else state
                for state in states
            ]
        elif self.arch == "cnn":
            states = [state.unsqueeze(2) for state in states]

        action_dict = {}
        for idx, (agent_id, state, actor) in enumerate(zip(agent_ids, states, actors)):
            if random.random() < epsilon:
                if self.discrete_actions:
                    action = (
                        np.random.dirichlet(
                            np.ones(self.action_dims[idx]), state.size()[0]
                        )
                        .astype("float32")
                        .squeeze()
                    )
                else:
                    action = (
                        (self.max_action[idx][0] - self.min_action[idx][0])
                        * np.random.rand(state.size()[0], self.action_dims[idx])
                        .astype("float32")
                        .squeeze()
                    ) + self.min_action[idx][0]
            else:
                actor.eval()
                if self.accelerator is not None:
                    with actor.no_sync(), torch.no_grad():
                        action_values = actor(state)
                else:
                    with torch.no_grad():
                        action_values = actor(state)
                actor.train()
                if self.discrete_actions:
                    action = action_values.cpu().data.numpy().squeeze()
                else:
                    action = self.scale_to_action_space(
                        action_values.cpu().data.numpy().squeeze(), idx
                    )
                    action = (
                        action
                        + np.random.normal(
                            0, self.expl_noise, size=self.action_dims[idx]
                        ).astype(np.float32)
                    ).clip(self.min_action[idx][0], self.max_action[idx][0])
            action_dict[agent_id] = action

        if self.discrete_actions:
            discrete_action_dict = {}
            for agent, action in action_dict.items():
                if self.one_hot:
                    discrete_action_dict[agent] = action.argmax(axis=1)
                else:
                    discrete_action_dict[agent] = action.argmax().item()
        else:
            discrete_action_dict = None

        if env_defined_actions is not None:
            for agent in env_defined_actions.keys():
                if not agent_mask[agent]:
                    if self.discrete_actions:
                        discrete_action_dict.update({agent: env_defined_actions[agent]})
                        action = env_defined_actions[agent]
                        action = (
                            nn.functional.one_hot(
                                torch.tensor(action).long(),
                                num_classes=action_dims_dict[agent],
                            )
                            .float()
                            .squeeze()
                        )
                        action_dict.update({agent: action})
                    else:
                        action_dict.update({agent: env_defined_actions[agent]})

        return (
            action_dict,
            discrete_action_dict,
        )

    def learn(self, experiences):
        """Updates agent network parameters to learn from experiences.

        :param experience: Tuple of dictionaries containing batched states, actions, rewards, next_states,
        dones in that order for each individual agent.
        :type experience: Tuple[Dict[str, torch.Tensor]]
        """

        for idx, (
            agent_id,
            actor,
            actor_target,
            critic_1,
            critic_target_1,
            critic_2,
            critic_target_2,
            actor_optimizer,
            critic_1_optimizer,
            critic_2_optimizer,
        ) in enumerate(
            zip(
                self.agent_ids,
                self.actors,
                self.actor_targets,
                self.critics_1,
                self.critic_targets_1,
                self.critics_2,
                self.critic_targets_2,
                self.actor_optimizers,
                self.critic_1_optimizers,
                self.critic_2_optimizers,
            )
        ):
            states, actions, rewards, next_states, dones = experiences

            if self.one_hot:
                states = {
                    agent_id: nn.functional.one_hot(
                        state.long(), num_classes=state_dim[0]
                    )
                    .float()
                    .squeeze()
                    for (agent_id, state), state_dim in zip(
                        states.items(), self.state_dims
                    )
                }
                next_states = {
                    agent_id: nn.functional.one_hot(
                        next_state.long(), num_classes=state_dim[0]
                    )
                    .float()
                    .squeeze()
                    for (agent_id, next_state), state_dim in zip(
                        next_states.items(), self.state_dims
                    )
                }

            if self.arch == "mlp":
                action_values = list(actions.values())
                input_combined = torch.cat(list(states.values()) + action_values, 1)
                if self.accelerator is not None:
                    with critic_1.no_sync():
                        q_value_1 = critic_1(input_combined)
                    with critic_2.no_sync():
                        q_value_2 = critic_2(input_combined)
                else:
                    q_value_1 = critic_1(input_combined)
                    q_value_2 = critic_2(input_combined)
                next_actions = []
                for i, agent_id_label in enumerate(self.agent_ids):
                    unscaled_actions = self.actor_targets[i](
                        next_states[agent_id_label]
                    ).detach_()
                    if not self.discrete_actions:
                        scaled_actions = torch.where(
                            unscaled_actions > 0,
                            unscaled_actions * self.max_action[i][0],
                            unscaled_actions * -self.min_action[i][0],
                        )
                        next_actions.append(scaled_actions)
                    else:
                        next_actions.append(unscaled_actions)

            elif self.arch == "cnn":
                stacked_states = torch.stack(list(states.values()), dim=2)
                stacked_actions = torch.cat(list(actions.values()), dim=1)
                if self.accelerator is not None:
                    with critic_1.no_sync():
                        q_value_1 = critic_1(stacked_states, stacked_actions)
                    with critic_2.no_sync():
                        q_value_2 = critic_2(stacked_states, stacked_actions)
                else:
                    q_value_1 = critic_1(stacked_states, stacked_actions)
                    q_value_2 = critic_2(stacked_states, stacked_actions)
                next_actions = []
                for i, agent_id_label in enumerate(self.agent_ids):
                    unscaled_actions = self.actor_targets[i](
                        next_states[agent_id_label].unsqueeze(2)
                    ).detach_()
                    if not self.discrete_actions:
                        scaled_actions = torch.where(
                            unscaled_actions > 0,
                            unscaled_actions * self.max_action[i][0],
                            unscaled_actions * -self.min_action[i][0],
                        )
                        next_actions.append(scaled_actions)
                    else:
                        next_actions.append(unscaled_actions)

            if self.arch == "mlp":
                next_input_combined = torch.cat(
                    list(next_states.values()) + next_actions, 1
                )
                if self.accelerator is not None:
                    with critic_target_1.no_sync():
                        q_value_next_state_1 = critic_target_1(next_input_combined)
                    with critic_target_2.no_sync():
                        q_value_next_state_2 = critic_target_2(next_input_combined)
                else:
                    q_value_next_state_1 = critic_target_1(next_input_combined)
                    q_value_next_state_2 = critic_target_2(next_input_combined)
            elif self.arch == "cnn":
                stacked_next_states = torch.stack(list(next_states.values()), dim=2)
                stacked_next_actions = torch.cat(next_actions, dim=1)
                if self.accelerator is not None:
                    with critic_target_1.no_sync():
                        q_value_next_state_1 = critic_target_1(
                            stacked_next_states, stacked_next_actions
                        )
                    with critic_target_2.no_sync():
                        q_value_next_state_2 = critic_target_2(
                            stacked_next_states, stacked_next_actions
                        )
                else:
                    q_value_next_state_1 = critic_target_1(
                        stacked_next_states, stacked_next_actions
                    )
                    q_value_next_state_2 = critic_target_2(
                        stacked_next_states, stacked_next_actions
                    )
            q_value_next_state = torch.min(q_value_next_state_1, q_value_next_state_2)

            y_j = (
                rewards[agent_id]
                + (1 - dones[agent_id]) * self.gamma * q_value_next_state
            )

            critic_loss = self.criterion(q_value_1, y_j.detach_()) + self.criterion(
                q_value_2, y_j.detach_()
            )

            # critic loss backprop
            critic_1_optimizer.zero_grad()
            critic_2_optimizer.zero_grad()
            if self.accelerator is not None:
                self.accelerator.backward(critic_loss)
            else:
                critic_loss.backward()
            critic_1_optimizer.step()
            critic_2_optimizer.step()

            # update actor and targets every policy_freq episodes
            if len(self.scores) % self.policy_freq == 0:
                if self.arch == "mlp":
                    if self.accelerator is not None:
                        with actor.no_sync():
                            action = actor(states[agent_id])
                    else:
                        action = actor(states[agent_id])
                    if not self.discrete_actions:
                        action = torch.where(
                            action > 0,
                            action * self.max_action[idx][0],
                            action * -self.min_action[idx][0],
                        )
                    detached_actions = copy.deepcopy(actions)
                    detached_actions[agent_id] = action
                    input_combined = torch.cat(
                        list(states.values()) + list(detached_actions.values()), 1
                    )
                    if self.accelerator is not None:
                        with critic_1.no_sync():
                            actor_loss = -critic_1(input_combined).mean()
                    else:
                        actor_loss = -critic_1(input_combined).mean()

                elif self.arch == "cnn":
                    if self.accelerator is not None:
                        with actor.no_sync():
                            action = actor(states[agent_id].unsqueeze(2))
                    else:
                        action = actor(states[agent_id].unsqueeze(2))
                    if not self.discrete_actions:
                        action = torch.where(
                            action > 0,
                            action * self.max_action[idx][0],
                            action * -self.min_action[idx][0],
                        )
                    detached_actions = copy.deepcopy(actions)
                    detached_actions[agent_id] = action
                    stacked_detached_actions = torch.cat(
                        list(detached_actions.values()), dim=1
                    )
                    if self.accelerator is not None:
                        with critic_1.no_sync():
                            actor_loss = -critic_1(
                                stacked_states, stacked_detached_actions
                            ).mean()
                    else:
                        actor_loss = -critic_1(
                            stacked_states, stacked_detached_actions
                        ).mean()

                # actor loss backprop
                actor_optimizer.zero_grad()
                if self.accelerator is not None:
                    self.accelerator.backward(actor_loss)
                else:
                    actor_loss.backward()
                actor_optimizer.step()

        if len(self.scores) % self.policy_freq == 0:
            for (
                actor,
                actor_target,
                critic_1,
                critic_target_1,
                critic_2,
                critic_target_2,
            ) in zip(
                self.actors,
                self.actor_targets,
                self.critics_1,
                self.critic_targets_1,
                self.critics_2,
                self.critic_targets_2,
            ):
                self.softUpdate(actor, actor_target)
                self.softUpdate(critic_1, critic_target_1)
                self.softUpdate(critic_2, critic_target_2)

            return actor_loss.item(), critic_loss.item()
        else:
            return None, critic_loss.item()

    def softUpdate(self, net, target):
        """Soft updates target network."""
        for eval_param, target_param in zip(net.parameters(), target.parameters()):
            target_param.data.copy_(
                self.tau * eval_param.data + (1.0 - self.tau) * target_param.data
            )

    def test(self, env, swap_channels=False, max_steps=500, loop=3):
        """Returns mean test score of agent in environment with epsilon-greedy policy.

        :param env: The environment to be tested in
        :type env: Gym-style environment
        :param swap_channels: Swap image channels dimension from last to first [H, W, C] -> [C, H, W], defaults to False
        :type swap_channels: bool, optional
        :param max_steps: Maximum number of testing steps, defaults to 500
        :type max_steps: int, optional
        :param loop: Number of testing loops/episodes to complete. The returned score is the mean. Defaults to 3
        :type loop: int, optional
        """
        with torch.no_grad():
            rewards = []
            for i in range(loop):
                state, info = env.reset()
                agent_reward = {agent_id: 0 for agent_id in self.agent_ids}
                score = 0
                for _ in range(max_steps):
                    if swap_channels:
                        state = {
                            agent_id: np.moveaxis(np.expand_dims(s, 0), [3], [1])
                            for agent_id, s in state.items()
                        }
                    agent_mask = (
                        info["agent_mask"] if "agent_mask" in info.keys() else None
                    )
                    env_defined_actions = (
                        info["env_defined_actions"]
                        if "env_defined_actions" in info.keys()
                        else None
                    )
                    cont_actions, discrete_action = self.getAction(
                        state,
                        epsilon=0,
                        agent_mask=agent_mask,
                        env_defined_actions=env_defined_actions,
                    )
                    if self.discrete_actions:
                        action = discrete_action
                    else:
                        action = cont_actions
                    state, reward, done, trunc, info = env.step(action)
                    for agent_id, r in reward.items():
                        agent_reward[agent_id] += r
                    score = sum(agent_reward.values())
                rewards.append(score)
        mean_fit = np.mean(rewards)
        self.fitness.append(mean_fit)
        return mean_fit

    def clone(self, index=None, wrap=True):
        """Returns cloned agent identical to self.

        :param index: Index to keep track of agent for tournament selection and mutation, defaults to None
        :type index: int, optional
        """
        if index is None:
            index = self.index

        clone = type(self)(
            state_dims=self.state_dims,
            action_dims=self.action_dims,
            one_hot=self.one_hot,
            n_agents=self.n_agents,
            agent_ids=self.agent_ids,
            max_action=self.max_action,
            min_action=self.min_action,
            expl_noise=self.expl_noise,
            discrete_actions=self.discrete_actions,
            index=index,
            net_config=self.net_config,
            batch_size=self.batch_size,
            policy_freq=self.policy_freq,
            lr=self.lr,
            learn_step=self.learn_step,
            gamma=self.gamma,
            tau=self.tau,
            mutation=self.mut,
            actor_networks=self.actor_networks,
            critic_networks=self.critic_networks,
            device=self.device,
            accelerator=self.accelerator,
            wrap=wrap,
        )

        if self.accelerator is not None:
            self.unwrap_models()
        actors = [actor.clone() for actor in self.actors]
        actor_targets = [actor_target.clone() for actor_target in self.actor_targets]
        critics_1 = [critic.clone() for critic in self.critics_1]
        critic_targets_1 = [
            critic_target.clone() for critic_target in self.critic_targets_1
        ]
        critics_2 = [critic.clone() for critic in self.critics_2]
        critic_targets_2 = [
            critic_target.clone() for critic_target in self.critic_targets_2
        ]
        actor_optimizers = [
            optim.Adam(actor.parameters(), lr=clone.lr) for actor in actors
        ]
        critic_1_optimizers = [
            optim.Adam(critic.parameters(), lr=clone.lr) for critic in critics_1
        ]
        critic_2_optimizers = [
            optim.Adam(critic.parameters(), lr=clone.lr) for critic in critics_2
        ]
        clone.actor_optimizers_type = actor_optimizers
        clone.critic_1_optimizers_type = critic_1_optimizers
        clone.critic_2_optimizers_type = critic_2_optimizers

        if self.accelerator is not None:
            if wrap:
                clone.actors = [self.accelerator.prepare(actor) for actor in actors]
                clone.actor_targets = [
                    self.accelerator.prepare(actor_target)
                    for actor_target in actor_targets
                ]
                clone.critics_1 = [
                    self.accelerator.prepare(critic) for critic in critics_1
                ]
                clone.critic_targets_1 = [
                    self.accelerator.prepare(critic_target)
                    for critic_target in critic_targets_1
                ]
                clone.critics_2 = [
                    self.accelerator.prepare(critic) for critic in critics_2
                ]
                clone.critic_targets_2 = [
                    self.accelerator.prepare(critic_target)
                    for critic_target in critic_targets_2
                ]
                clone.actor_optimizers = [
                    self.accelerator.prepare(actor_optimizer)
                    for actor_optimizer in actor_optimizers
                ]
                clone.critic_1_optimizers = [
                    self.accelerator.prepare(critic_optimizer)
                    for critic_optimizer in critic_1_optimizers
                ]
                clone.critic_2_optimizers = [
                    self.accelerator.prepare(critic_optimizer)
                    for critic_optimizer in critic_2_optimizers
                ]
            else:
                (
                    clone.actors,
                    clone.actor_targets,
                    clone.critics_1,
                    clone.critic_targets_1,
                    clone.critics_2,
                    clone.critic_targets_2,
                    clone.actor_optimizers,
                    clone.critic_1_optimizers,
                    clone.critic_2_optimizers,
                ) = (
                    actors,
                    actor_targets,
                    critics_1,
                    critic_targets_1,
                    critics_2,
                    critic_targets_2,
                    actor_optimizers,
                    critic_1_optimizers,
                    critic_2_optimizers,
                )
        else:
            clone.actors = [actor.to(self.device) for actor in actors]
            clone.actor_targets = [
                actor_target.to(self.device) for actor_target in actor_targets
            ]
            clone.critics_1 = [critic.to(self.device) for critic in critics_1]
            clone.critic_targets_1 = [
                critic_target.to(self.device) for critic_target in critic_targets_1
            ]
            clone.critics_2 = [critic.to(self.device) for critic in critics_2]
            clone.critic_targets_2 = [
                critic_target.to(self.device) for critic_target in critic_targets_2
            ]
            clone.actor_optimizers = actor_optimizers
            clone.critic_1_optimizers = critic_1_optimizers
            clone.critic_2_optimizers = critic_2_optimizers

        clone.fitness = copy.deepcopy(self.fitness)
        clone.steps = copy.deepcopy(self.steps)
        clone.scores = copy.deepcopy(self.scores)

        return clone

    def wrap_models(self):
        if self.accelerator is not None:
            self.actors = [self.accelerator.prepare(actor) for actor in self.actors]
            self.actor_targets = [
                self.accelerator.prepare(actor_target)
                for actor_target in self.actor_targets
            ]
            self.critics_1 = [
                self.accelerator.prepare(critic) for critic in self.critics_1
            ]
            self.critic_targets_1 = [
                self.accelerator.prepare(critic_target)
                for critic_target in self.critic_targets_1
            ]
            self.critics_2 = [
                self.accelerator.prepare(critic) for critic in self.critics_2
            ]
            self.critic_targets_2 = [
                self.accelerator.prepare(critic_target)
                for critic_target in self.critic_targets_2
            ]
            self.actor_optimizers = [
                self.accelerator.prepare(actor_optimizer)
                for actor_optimizer in self.actor_optimizers_type
            ]
            self.critic_1_optimizers = [
                self.accelerator.prepare(critic_optimizer)
                for critic_optimizer in self.critic_1_optimizers_type
            ]
            self.critic_2_optimizers = [
                self.accelerator.prepare(critic_optimizer)
                for critic_optimizer in self.critic_2_optimizers_type
            ]

    def unwrap_models(self):
        if self.accelerator is not None:
            self.actors = [
                self.accelerator.unwrap_model(actor) for actor in self.actors
            ]
            self.actor_targets = [
                self.accelerator.unwrap_model(actor_target)
                for actor_target in self.actor_targets
            ]
            self.critics_1 = [
                self.accelerator.unwrap_model(critic) for critic in self.critics_1
            ]
            self.critic_targets_1 = [
                self.accelerator.unwrap_model(critic_target)
                for critic_target in self.critic_targets_1
            ]
            self.critics_2 = [
                self.accelerator.unwrap_model(critic) for critic in self.critics_2
            ]
            self.critic_targets_2 = [
                self.accelerator.unwrap_model(critic_target)
                for critic_target in self.critic_targets_2
            ]
            self.actor_optimizers = [
                self.accelerator.unwrap_model(actor_optimizer)
                for actor_optimizer in self.actor_optimizers
            ]
            self.critic_1_optimizers = [
                self.accelerator.unwrap_model(critic_optimizer)
                for critic_optimizer in self.critic_1_optimizers
            ]
            self.critic_2_optimizers = [
                self.accelerator.unwrap_model(critic_optimizer)
                for critic_optimizer in self.critic_2_optimizers
            ]

    def saveCheckpoint(self, path):
        """Saves a checkpoint of agent properties and network weights to path.

        :param path: Location to save checkpoint at
        :type path: string
        """

        torch.save(
            {
                "actors_init_dict": [actor.init_dict for actor in self.actors],
                "actors_state_dict": [actor.state_dict() for actor in self.actors],
                "actor_targets_init_dict": [
                    actor_target.init_dict for actor_target in self.actor_targets
                ],
                "actor_targets_state_dict": [
                    actor_target.state_dict() for actor_target in self.actor_targets
                ],
                "critics_1_init_dict": [
                    critic_1.init_dict for critic_1 in self.critics_1
                ],
                "critics_1_state_dict": [
                    critic_1.state_dict() for critic_1 in self.critics_1
                ],
                "critic_targets_1_init_dict": [
                    critic_target_1.init_dict
                    for critic_target_1 in self.critic_targets_1
                ],
                "critic_targets_1_state_dict": [
                    critic_target_1.state_dict()
                    for critic_target_1 in self.critic_targets_1
                ],
                "critics_2_init_dict": [
                    critic_2.init_dict for critic_2 in self.critics_2
                ],
                "critics_2_state_dict": [
                    critic_2.state_dict() for critic_2 in self.critics_2
                ],
                "critic_targets_2_init_dict": [
                    critic_target_2.init_dict
                    for critic_target_2 in self.critic_targets_2
                ],
                "critic_targets_2_state_dict": [
                    critic_target_2.state_dict()
                    for critic_target_2 in self.critic_targets_2
                ],
                "actor_optimizers_state_dict": [
                    actor_optimizer.state_dict()
                    for actor_optimizer in self.actor_optimizers
                ],
                "critic_1_optimizers_state_dict": [
                    critic_1_optimizer.state_dict()
                    for critic_1_optimizer in self.critic_1_optimizers
                ],
                "critic_2_optimizers_state_dict": [
                    critic_2_optimizer.state_dict()
                    for critic_2_optimizer in self.critic_2_optimizers
                ],
                "net_config": self.net_config,
                "batch_size": self.batch_size,
                "lr": self.lr,
                "learn_step": self.learn_step,
                "policy_freq": self.policy_freq,
                "gamma": self.gamma,
                "tau": self.tau,
                "mutation": self.mut,
                "index": self.index,
                "scores": self.scores,
                "fitness": self.fitness,
                "steps": self.steps,
            },
            path,
            pickle_module=dill,
        )

    def loadCheckpoint(self, path):
        """Loads saved agent properties and network weights from checkpoint.

        :param path: Location to load checkpoint from
        :type path: string
        """
        checkpoint = torch.load(path, pickle_module=dill)
        self.net_config = checkpoint["net_config"]
        if self.net_config is not None:
            self.arch = checkpoint["net_config"]["arch"]
            if self.arch == "mlp":
                self.actors = [
                    EvolvableMLP(**checkpoint["actors_init_dict"][idx])
                    for idx, _ in enumerate(self.agent_ids)
                ]
                self.actor_targets = [
                    EvolvableMLP(**checkpoint["actor_targets_init_dict"][idx])
                    for idx, _ in enumerate(self.agent_ids)
                ]
                self.critics_1 = [
                    EvolvableMLP(**checkpoint["critics_1_init_dict"][idx])
                    for idx, _ in enumerate(self.agent_ids)
                ]
                self.critic_targets_1 = [
                    EvolvableMLP(**checkpoint["critic_targets_1_init_dict"][idx])
                    for idx, _ in enumerate(self.agent_ids)
                ]
                self.critics_2 = [
                    EvolvableMLP(**checkpoint["critics_2_init_dict"][idx])
                    for idx, _ in enumerate(self.agent_ids)
                ]
                self.critic_targets_2 = [
                    EvolvableMLP(**checkpoint["critic_targets_2_init_dict"][idx])
                    for idx, _ in enumerate(self.agent_ids)
                ]
            elif self.arch == "cnn":
                self.actors = [
                    EvolvableCNN(**checkpoint["actors_init_dict"][idx])
                    for idx, _ in enumerate(self.agent_ids)
                ]
                self.actor_targets = [
                    EvolvableCNN(**checkpoint["actor_targets_init_dict"][idx])
                    for idx, _ in enumerate(self.agent_ids)
                ]
                self.critics_1 = [
                    EvolvableCNN(**checkpoint["critics_1_init_dict"][idx])
                    for idx, _ in enumerate(self.agent_ids)
                ]
                self.critic_targets_1 = [
                    EvolvableCNN(**checkpoint["critic_targets_1_init_dict"][idx])
                    for idx, _ in enumerate(self.agent_ids)
                ]
                self.critics_2 = [
                    EvolvableCNN(**checkpoint["critics_2_init_dict"][idx])
                    for idx, _ in enumerate(self.agent_ids)
                ]
                self.critic_targets_2 = [
                    EvolvableCNN(**checkpoint["critic_targets_2_init_dict"][idx])
                    for idx, _ in enumerate(self.agent_ids)
                ]
        else:
            self.actors = [
                MakeEvolvable(**checkpoint["actors_init_dict"][idx])
                for idx, _ in enumerate(self.agent_ids)
            ]
            self.actor_targets = [
                MakeEvolvable(**checkpoint["actor_targets_init_dict"][idx])
                for idx, _ in enumerate(self.agent_ids)
            ]
            self.critics_1 = [
                MakeEvolvable(**checkpoint["critics_1_init_dict"][idx])
                for idx, _ in enumerate(self.agent_ids)
            ]
            self.critic_targets_1 = [
                MakeEvolvable(**checkpoint["critic_targets_1_init_dict"][idx])
                for idx, _ in enumerate(self.agent_ids)
            ]
            self.critics_2 = [
                MakeEvolvable(**checkpoint["critics_2_init_dict"][idx])
                for idx, _ in enumerate(self.agent_ids)
            ]
            self.critic_targets_2 = [
                MakeEvolvable(**checkpoint["critic_targets_2_init_dict"][idx])
                for idx, _ in enumerate(self.agent_ids)
            ]
        self.lr = checkpoint["lr"]
        self.actor_optimizers = [
            optim.Adam(actor.parameters(), lr=self.lr) for actor in self.actors
        ]
        self.critic_1_optimizers = [
            optim.Adam(critic_1.parameters(), lr=self.lr) for critic_1 in self.critics_1
        ]
        self.critic_2_optimizers = [
            optim.Adam(critic_2.parameters(), lr=self.lr) for critic_2 in self.critics_2
        ]
        actor_list = []
        critic_1_list = []
        critic_2_list = []
        actor_target_list = []
        critic_target_1_list = []
        critic_target_2_list = []
        actor_optimizer_list = []
        critic_1_optimizer_list = []
        critic_2_optimizer_list = []

        for idx, (
            actor,
            actor_target,
            critic_1,
            critic_target_1,
            critic_2,
            critic_target_2,
            actor_optimizer,
            critic_1_optimizer,
            critic_2_optimizer,
        ) in enumerate(
            zip(
                self.actors,
                self.actor_targets,
                self.critics_1,
                self.critic_targets_1,
                self.critics_2,
                self.critic_targets_2,
                self.actor_optimizers,
                self.critic_1_optimizers,
                self.critic_2_optimizers,
            )
        ):
            actor.load_state_dict(checkpoint["actors_state_dict"][idx])
            actor_list.append(actor)
            actor_target.load_state_dict(checkpoint["actor_targets_state_dict"][idx])
            actor_target_list.append(actor_target)
            critic_1.load_state_dict(checkpoint["critics_1_state_dict"][idx])
            critic_1_list.append(critic_1)
            critic_2.load_state_dict(checkpoint["critics_2_state_dict"][idx])
            critic_2_list.append(critic_2)
            critic_target_1.load_state_dict(
                checkpoint["critic_targets_1_state_dict"][idx]
            )
            critic_target_1_list.append(critic_target_1)
            critic_target_2.load_state_dict(
                checkpoint["critic_targets_2_state_dict"][idx]
            )
            critic_target_2_list.append(critic_target_2)
            actor_optimizer.load_state_dict(
                checkpoint["actor_optimizers_state_dict"][idx]
            )
            actor_optimizer_list.append(actor_optimizer)
            critic_1_optimizer.load_state_dict(
                checkpoint["critic_1_optimizers_state_dict"][idx]
            )
            critic_1_optimizer_list.append(critic_1_optimizer)
            critic_2_optimizer.load_state_dict(
                checkpoint["critic_2_optimizers_state_dict"][idx]
            )
            critic_2_optimizer_list.append(critic_2_optimizer)

        self.actors = actor_list
        self.actor_targets = actor_target_list
        self.critics_1 = critic_1_list
        self.critic_targets_1 = critic_target_1_list
        self.critics_2 = critic_2_list
        self.critic_targets_2 = critic_target_2_list
        self.actor_optimizers = actor_optimizer_list
        self.critic_1_optimizers = critic_1_optimizer_list
        self.critic_2_optimizers = critic_2_optimizer_list
        self.batch_size = checkpoint["batch_size"]
        self.learn_step = checkpoint["learn_step"]
        self.policy_freq = checkpoint["policy_freq"]
        self.gamma = checkpoint["gamma"]
        self.tau = checkpoint["tau"]
        self.mut = checkpoint["mutation"]
        self.index = checkpoint["index"]
        self.scores = checkpoint["scores"]
        self.fitness = checkpoint["fitness"]
        self.steps = checkpoint["steps"]



import copy

import dill
import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
from torch.distributions import Categorical, MultivariateNormal

from agilerl.networks.evolvable_cnn import EvolvableCNN
from agilerl.networks.evolvable_mlp import EvolvableMLP
from agilerl.wrappers.make_evolvable import MakeEvolvable


class PPO:
    """The PPO algorithm class. PPO paper: https://arxiv.org/abs/1707.06347v2

    :param state_dim: State observation dimension
    :type state_dim: list[int]
    :param action_dim: Action dimension
    :type action_dim: int
    :param one_hot: One-hot encoding, used with discrete observation spaces
    :type one_hot: bool
    :param discrete_actions: Boolean flag to indicate a discrete action space
    :type discrete_actions: bool, optional
    :param index: Index to keep track of object instance during tournament selection and mutation, defaults to 0
    :type index: int, optional
    :param net_config: Network configuration, defaults to mlp with hidden size [64,64]
    :type net_config: dict, optional
    :param batch_size: Size of batched sample from replay buffer for learning, defaults to 64
    :type batch_size: int, optional
    :param lr: Learning rate for optimizer, defaults to 1e-4
    :type lr: float, optional
    :param gamma: Discount factor, defaults to 0.99
    :type gamma: float, optional
    :param gae_lambda: Lambda for general advantage estimation, defaults to 0.95
    :type gae_lambda: float, optional
    :param mutation: Most recent mutation to agent, defaults to None
    :type mutation: str, optional
    :param action_std_init: Initial action standard deviation, defaults to 0.6
    :type action_std_init: float, optional
    :param clip_coef: Surrogate clipping coefficient, defaults to 0.2
    :type clip coef: float, optional
    :param ent_coef: Entropy coefficient, defaults to 0.01
    :type ent_coef: float, optional
    :param vf_coef: Value function coefficient, defaults to 0.5
    :type vf_coef: float, optional
    :param max_grad_norm: Maximum norm for gradient clipping, defaults to 0.5
    :type max_grad_norm: float, optional
    :param target_kl: Target KL divergence threshold, defaults to None
    :type target_kl: float, optional
    :param update_epochs: Number of policy update epochs, defaults to 4
    :type update_epochs: int, optional
    :param actor_network: Custom actor network, defaults to None
    :type actor_network: nn.Module, optional
    :param critic_network: Custom critic network, defaults to None
    :type critic_network: nn.Module, optional
    :param device: Device for accelerated computing, 'cpu' or 'cuda', defaults to 'cpu'
    :type device: str, optional
    :param accelerator: Accelerator for distributed computing, defaults to None
    :type accelerator: accelerate.Accelerator(), optional
    :param wrap: Wrap models for distributed training upon creation, defaults to True
    :type wrap: bool, optional
    """

    def __init__(
        self,
        state_dim,
        action_dim,
        one_hot,
        discrete_actions,
        index=0,
        net_config={"arch": "mlp", "h_size": [64, 64]},
        batch_size=64,
        lr=1e-4,
        gamma=0.99,
        gae_lambda=0.95,
        mutation=None,
        action_std_init=0.6,
        clip_coef=0.2,
        ent_coef=0.01,
        vf_coef=0.5,
        max_grad_norm=0.5,
        target_kl=None,
        update_epochs=4,
        actor_network=None,
        critic_network=None,
        device="cpu",
        accelerator=None,
        wrap=True,
    ):
        assert isinstance(
            state_dim, (list, tuple)
        ), "State dimension must be a list or tuple."
        assert isinstance(
            action_dim, (int, np.integer)
        ), "Action dimension must be an integer."
        assert isinstance(
            one_hot, bool
        ), "One-hot encoding flag must be boolean value True or False."
        assert isinstance(
            discrete_actions, bool
        ), "Discrete actions flag must be boolean value True or False."
        assert isinstance(index, int), "Agent index must be an integer."
        assert isinstance(batch_size, int), "Batch size must be an integer."
        assert batch_size >= 1, "Batch size must be greater than or equal to one."
        assert isinstance(lr, float), "Learning rate must be a float."
        assert lr > 0, "Learning rate must be greater than zero."
        assert isinstance(gamma, (float, int)), "Gamma must be a float."
        assert isinstance(gae_lambda, (float, int)), "Lambda must be a float."
        assert gae_lambda >= 0, "Lambda must be greater than or equal to zero."
        assert isinstance(
            action_std_init, (float, int)
        ), "Action standard deviation must be a float."
        assert (
            action_std_init >= 0
        ), "Action standard deviation must be greater than or equal to zero."
        assert isinstance(
            clip_coef, (float, int)
        ), "Clipping coefficient must be a float."
        assert (
            clip_coef >= 0
        ), "Clipping coefficient must be greater than or equal to zero."
        assert isinstance(
            ent_coef, (float, int)
        ), "Entropy coefficient must be a float."
        assert (
            ent_coef >= 0
        ), "Entropy coefficient must be greater than or equal to zero."
        assert isinstance(
            vf_coef, (float, int)
        ), "Value function coefficient must be a float."
        assert (
            vf_coef >= 0
        ), "Value function coefficient must be greater than or equal to zero."
        assert isinstance(
            max_grad_norm, (float, int)
        ), "Maximum norm for gradient clipping must be a float."
        assert (
            max_grad_norm >= 0
        ), "Maximum norm for gradient clipping must be greater than or equal to zero."
        assert (
            isinstance(target_kl, (float, int)) or target_kl is None
        ), "Target KL divergence threshold must be a float."
        if target_kl is not None:
            assert (
                target_kl >= 0
            ), "Target KL divergence threshold must be greater than or equal to zero."
        assert isinstance(
            update_epochs, int
        ), "Policy update epochs must be an integer."
        assert (
            update_epochs >= 1
        ), "Policy update epochs must be greater than or equal to one."
        assert (
            isinstance(actor_network, nn.Module) or actor_network is None
        ), "Actor network must be an nn.Module or None."
        assert isinstance(
            wrap, bool
        ), "Wrap models flag must be boolean value True or False."

        self.algo = "PPO"
        self.state_dim = state_dim
        self.action_dim = action_dim
        self.one_hot = one_hot
        self.discrete_actions = discrete_actions
        self.net_config = net_config
        self.batch_size = batch_size
        self.lr = lr
        self.gamma = gamma
        self.gae_lambda = gae_lambda
        self.mut = mutation
        self.action_std_init = action_std_init
        self.clip_coef = clip_coef
        self.ent_coef = ent_coef
        self.vf_coef = vf_coef
        self.max_grad_norm = max_grad_norm
        self.target_kl = target_kl
        self.update_epochs = update_epochs
        self.actor_network = actor_network
        self.critic_network = critic_network
        self.device = device
        self.accelerator = accelerator

        self.index = index
        self.scores = []
        self.fitness = []
        self.steps = [0]

        # For continuous action spaces
        if not self.discrete_actions:
            self.action_var = torch.full((action_dim,), action_std_init**2)
            if self.accelerator is None:
                self.action_var = self.action_var.to(self.device)
            else:
                self.action_var = self.action_var.to(self.accelerator.device)

        if self.actor_network is not None and self.critic_network is not None:
            self.actor = actor_network
            self.critic = critic_network
            self.net_config = None
        else:
            assert isinstance(self.net_config, dict), "Net config must be a dictionary."
            assert (
                "arch" in self.net_config.keys()
            ), "Net config must contain arch: 'mlp' or 'cnn'."
            # Set up network output activations
            if "output_activation" in self.net_config.keys():
                pass
            else:
                if self.discrete_actions:
                    self.net_config["output_activation"] = "Softmax"
                else:
                    self.net_config["output_activation"] = "Tanh"
            # model
            if self.net_config["arch"] == "mlp":  # Multi-layer Perceptron
                assert (
                    "h_size" in self.net_config.keys()
                ), "Net config must contain h_size: int."
                assert isinstance(
                    self.net_config["h_size"], list
                ), "Net config h_size must be a list."
                assert (
                    len(self.net_config["h_size"]) > 0
                ), "Net config h_size must contain at least one element."
                self.actor = EvolvableMLP(
                    num_inputs=state_dim[0],
                    num_outputs=action_dim,
                    hidden_size=self.net_config["h_size"],
                    mlp_output_activation=self.net_config["output_activation"],
                    device=self.device,
                    accelerator=self.accelerator,
                )
                self.critic = EvolvableMLP(
                    num_inputs=state_dim[0],
                    num_outputs=1,
                    hidden_size=self.net_config["h_size"],
                    device=self.device,
                    accelerator=self.accelerator,
                )
            elif self.net_config["arch"] == "cnn":  # Convolutional Neural Network
                for key in ["c_size", "k_size", "s_size", "h_size"]:
                    assert (
                        key in self.net_config.keys()
                    ), f"Net config must contain {key}: int."
                    assert isinstance(
                        self.net_config[key], list
                    ), f"Net config {key} must be a list."
                    assert (
                        len(self.net_config[key]) > 0
                    ), f"Net config {key} must contain at least one element."
                assert (
                    "normalize" in self.net_config.keys()
                ), "Net config must contain normalize: True or False."
                assert isinstance(
                    self.net_config["normalize"], bool
                ), "Net config normalize must be boolean value True or False."
                self.actor = EvolvableCNN(
                    input_shape=state_dim,
                    num_actions=action_dim,
                    channel_size=self.net_config["c_size"],
                    kernel_size=self.net_config["k_size"],
                    stride_size=self.net_config["s_size"],
                    hidden_size=self.net_config["h_size"],
                    normalize=self.net_config["normalize"],
                    mlp_output_activation=self.net_config["output_activation"],
                    device=self.device,
                    accelerator=self.accelerator,
                )
                self.critic = EvolvableCNN(
                    input_shape=state_dim,
                    num_actions=1,
                    channel_size=self.net_config["c_size"],
                    kernel_size=self.net_config["k_size"],
                    stride_size=self.net_config["s_size"],
                    hidden_size=self.net_config["h_size"],
                    normalize=self.net_config["normalize"],
                    mlp_activation="Tanh",
                    device=self.device,
                    accelerator=self.accelerator,
                )

        self.arch = (
            self.net_config["arch"] if self.net_config is not None else self.actor.arch
        )

        self.optimizer_type = optim.Adam(
            [
                {"params": self.actor.parameters(), "lr": self.lr},
                {"params": self.critic.parameters(), "lr": self.lr},
            ]
        )

        if self.accelerator is not None:
            self.optimizer = self.optimizer_type
            if wrap:
                self.wrap_models()
        else:
            self.actor = self.actor.to(self.device)
            self.critic = self.critic.to(self.device)
            self.optimizer = self.optimizer_type

        self.criterion = nn.MSELoss()

    def prepare_state(self, state):
        """Prepares state for forward pass through neural network.

        :param state: Observation of environment
        :type state: np.Array() or list
        """
        if not isinstance(state, torch.Tensor):
            state = torch.from_numpy(state).float()

        if self.accelerator is None:
            state = state.to(self.device)
        else:
            state = state.to(self.accelerator.device)

        if self.one_hot:
            state = (
                nn.functional.one_hot(state.long(), num_classes=self.state_dim[0])
                .float()
                .squeeze()
            )

        if len(state.size()) < 2:
            state = state.unsqueeze(0)

        if self.arch == "cnn" and state.dim() <= 3:
            state = state.unsqueeze(0)

        return state.float()

    def getAction(self, state, action=None, grad=False):
        """Returns the next action to take in the environment.

        :param state: Environment observation, or multiple observations in a batch
        :type state: float or list[float]
        :param action: Action in environment to evaluate, defaults to None
        :type action: torch.Tensor(), optional
        :param grad: Calculate gradients on actions, defaults to False
        :type grad: bool, optional
        """
        state = self.prepare_state(state)

        if not grad:
            self.actor.eval()
            self.critic.eval()
            with torch.no_grad():
                action_values = self.actor(state)
                state_values = self.critic(state).squeeze(-1)
            self.actor.train()
            self.critic.train()

        else:
            action_values = self.actor(state)
            state_values = self.critic(state).squeeze(-1)

        if self.discrete_actions:
            dist = Categorical(action_values)
        else:
            cov_mat = torch.diag(self.action_var).unsqueeze(dim=0)
            dist = MultivariateNormal(action_values, cov_mat)

        return_tensors = True
        if action is None:
            action = dist.sample()
            return_tensors = False
        elif self.accelerator is None:
            action = action.to(self.device)
        else:
            action = action.to(self.accelerator.device)

        action_logprob = dist.log_prob(action)
        dist_entropy = dist.entropy()

        if return_tensors:
            return action, action_logprob, dist_entropy, state_values
        else:
            return (
                action.cpu().data.numpy(),
                action_logprob.cpu().data.numpy(),
                dist_entropy.cpu().data.numpy(),
                state_values.cpu().data.numpy(),
            )

    def learn(self, experiences, noise_clip=0.5, policy_noise=0.2):
        """Updates agent network parameters to learn from experiences.

        :param experience: List of batched states, actions, log_probs, rewards, dones, values, next_state in that order.
        :type experience: list[torch.Tensor[float]]
        :param noise_clip: Maximum noise limit to apply to actions, defaults to 0.5
        :type noise_clip: float, optional
        :param policy_noise: Standard deviation of noise applied to policy, defaults to 0.2
        :type policy_noise: float, optional
        """
        experiences = [torch.from_numpy(np.array(exp)) for exp in experiences]
        states, actions, log_probs, rewards, dones, values, next_state = experiences
        if self.accelerator is not None:
            next_state = next_state.to(self.accelerator.device)
        dones = dones.long()

        # Bootstrapping
        with torch.no_grad():
            num_steps = rewards.size(0)
            next_state = self.prepare_state(next_state)
            next_value = self.critic(next_state).reshape(1, -1).cpu()
            advantages = torch.zeros_like(rewards)
            last_gae_lambda = 0
            for t in reversed(range(num_steps)):
                if t == num_steps - 1:
                    nextnonterminal = 1.0 - dones[-1]
                    nextvalues = next_value
                else:
                    nextnonterminal = 1.0 - dones[t + 1]
                    nextvalues = values[t + 1]
                delta = (
                    rewards[t] + self.gamma * nextvalues * nextnonterminal - values[t]
                )
                advantages[t] = last_gae_lambda = (
                    delta
                    + self.gamma * self.gae_lambda * nextnonterminal * last_gae_lambda
                )
            returns = advantages + values

        states = states.reshape((-1,) + self.state_dim)
        if self.discrete_actions:
            actions = actions.reshape(-1)
        else:
            actions = actions.reshape((-1, self.action_dim))
        log_probs = log_probs.reshape(-1)
        advantages = advantages.reshape(-1)
        returns = returns.reshape(-1)
        values = values.reshape(-1)

        if self.accelerator is None:
            states, actions, log_probs, advantages, returns, values = (
                states.to(self.device),
                actions.to(self.device),
                log_probs.to(self.device),
                advantages.to(self.device),
                returns.to(self.device),
                values.to(self.device),
            )
        else:
            states = states.to(self.accelerator.device)
            actions = actions.to(self.accelerator.device)
            log_probs = log_probs.to(self.accelerator.device)
            advantages = advantages.to(self.accelerator.device)
            returns = returns.to(self.accelerator.device)
            values = values.to(self.accelerator.device)

        num_samples = returns.size(0)
        batch_idxs = np.arange(num_samples)

        clipfracs = []

        mean_loss = 0
        for epoch in range(self.update_epochs):
            np.random.shuffle(batch_idxs)
            for start in range(0, num_samples, self.batch_size):
                minibatch_idxs = batch_idxs[start : start + self.batch_size]

                _, log_prob, entropy, value = self.getAction(
                    state=states[minibatch_idxs],
                    action=actions[minibatch_idxs],
                    grad=True,
                )

                logratio = log_prob - log_probs[minibatch_idxs]
                ratio = logratio.exp()

                with torch.no_grad():
                    approx_kl = ((ratio - 1) - logratio).mean()
                    clipfracs += [
                        ((ratio - 1.0).abs() > self.clip_coef).float().mean().item()
                    ]

                minibatch_advs = advantages[minibatch_idxs]
                minibatch_advs = (minibatch_advs - minibatch_advs.mean()) / (
                    minibatch_advs.std() + 1e-8
                )

                # Policy loss
                pg_loss1 = -minibatch_advs * ratio
                pg_loss2 = -minibatch_advs * torch.clamp(
                    ratio, 1 - self.clip_coef, 1 + self.clip_coef
                )
                pg_loss = torch.max(pg_loss1, pg_loss2).mean()

                # Value loss
                value = value.view(-1)
                v_loss_unclipped = (value - returns[minibatch_idxs]) ** 2
                v_clipped = values[minibatch_idxs] + torch.clamp(
                    value - values[minibatch_idxs], -self.clip_coef, self.clip_coef
                )
                v_loss_clipped = (v_clipped - returns[minibatch_idxs]) ** 2
                v_loss_max = torch.max(v_loss_unclipped, v_loss_clipped)
                v_loss = 0.5 * v_loss_max.mean()

                entropy_loss = entropy.mean()
                loss = pg_loss - self.ent_coef * entropy_loss + v_loss * self.vf_coef

                # actor loss backprop
                self.optimizer.zero_grad()
                if self.accelerator is not None:
                    self.accelerator.backward(loss)
                else:
                    loss.backward()
                self.optimizer.step()

                mean_loss += loss.item()

            if self.target_kl is not None:
                if approx_kl > self.target_kl:
                    break

        mean_loss /= num_samples * self.update_epochs
        return mean_loss

    def test(self, env, swap_channels=False, max_steps=500, loop=3):
        """Returns mean test score of agent in environment with epsilon-greedy policy.

        :param env: The environment to be tested in
        :type env: Gym-style environment
        :param swap_channels: Swap image channels dimension from last to first [H, W, C] -> [C, H, W], defaults to False
        :type swap_channels: bool, optional
        :param max_steps: Maximum number of testing steps, defaults to 500
        :type max_steps: int, optional
        :param loop: Number of testing loops/episodes to complete. The returned score is the mean. Defaults to 3
        :type loop: int, optional
        """
        with torch.no_grad():
            rewards = []
            for i in range(loop):
                state = env.reset()[0]
                score = 0
                for idx_step in range(max_steps):
                    if swap_channels:
                        if not hasattr(env, "num_envs"):
                            state = np.expand_dims(state, 0)
                        state = np.moveaxis(state, [3], [1])
                    action, _, _, _ = self.getAction(state)
                    if not hasattr(env, "num_envs"):
                        action = action[0]
                    state, reward, done, trunc, info = env.step(action)
                    if hasattr(env, "num_envs"):
                        done = done[0]
                        trunc = trunc[0]
                        reward = reward[0]
                    score += reward
                    if done or trunc:
                        break
                rewards.append(score)
        mean_fit = np.mean(rewards)
        self.fitness.append(mean_fit)
        return mean_fit

    def clone(self, index=None, wrap=True):
        """Returns cloned agent identical to self.

        :param index: Index to keep track of agent for tournament selection and mutation, defaults to None
        :type index: int, optional
        """
        if index is None:
            index = self.index

        clone = type(self)(
            state_dim=self.state_dim,
            action_dim=self.action_dim,
            one_hot=self.one_hot,
            discrete_actions=self.discrete_actions,
            index=index,
            net_config=self.net_config,
            batch_size=self.batch_size,
            lr=self.lr,
            gamma=self.gamma,
            gae_lambda=self.gae_lambda,
            action_std_init=self.action_std_init,
            clip_coef=self.clip_coef,
            ent_coef=self.ent_coef,
            vf_coef=self.vf_coef,
            max_grad_norm=self.max_grad_norm,
            target_kl=self.target_kl,
            update_epochs=self.update_epochs,
            actor_network=self.actor_network,
            critic_network=self.critic_network,
            mutation=self.mut,
            device=self.device,
            accelerator=self.accelerator,
            wrap=wrap,
        )

        if self.accelerator is not None:
            self.unwrap_models()
        actor = self.actor.clone()
        critic = self.critic.clone()
        optimizer = optim.Adam(
            [
                {"params": actor.parameters(), "lr": self.lr},
                {"params": critic.parameters(), "lr": self.lr},
            ]
        )
        clone.optimizer_type = optimizer

        if self.accelerator is not None:
            if wrap:
                (
                    clone.actor,
                    clone.critic,
                    clone.optimizer,
                ) = self.accelerator.prepare(actor, critic, optimizer)
            else:
                (
                    clone.actor,
                    clone.critic,
                    clone.optimizer,
                ) = (
                    actor,
                    critic,
                    optimizer,
                )
        else:
            clone.actor = actor.to(self.device)
            clone.critic = critic.to(self.device)
            clone.optimizer = optimizer

        clone.fitness = copy.deepcopy(self.fitness)
        clone.steps = copy.deepcopy(self.steps)
        clone.scores = copy.deepcopy(self.scores)

        return clone

    def wrap_models(self):
        if self.accelerator is not None:
            self.actor, self.critic, self.optimizer = self.accelerator.prepare(
                self.actor, self.critic, self.optimizer_type
            )

    def unwrap_models(self):
        if self.accelerator is not None:
            self.actor = self.accelerator.unwrap_model(self.actor)
            self.critic = self.accelerator.unwrap_model(self.critic)
            self.optimizer = self.accelerator.unwrap_model(self.optimizer)

    def saveCheckpoint(self, path):
        """Saves a checkpoint of agent properties and network weights to path.

        :param path: Location to save checkpoint at
        :type path: string
        """
        torch.save(
            {
                "actor_init_dict": self.actor.init_dict,
                "actor_state_dict": self.actor.state_dict(),
                "critic_init_dict": self.critic.init_dict,
                "critic_state_dict": self.critic.state_dict(),
                "optimizer_state_dict": self.optimizer.state_dict(),
                "discrete_actions": self.discrete_actions,
                "net_config": self.net_config,
                "batch_size": self.batch_size,
                "lr": self.lr,
                "gamma": self.gamma,
                "gae_lambda": self.gae_lambda,
                "action_std_init": self.action_std_init,
                "clip_coef": self.clip_coef,
                "ent_coef": self.ent_coef,
                "vf_coef": self.vf_coef,
                "max_grad_norm": self.max_grad_norm,
                "target_kl": self.target_kl,
                "update_epochs": self.update_epochs,
                "mutation": self.mut,
                "index": self.index,
                "scores": self.scores,
                "fitness": self.fitness,
                "steps": self.steps,
            },
            path,
            pickle_module=dill,
        )

    def loadCheckpoint(self, path):
        """Loads saved agent properties and network weights from checkpoint.

        :param path: Location to load checkpoint from
        :type path: string
        """
        checkpoint = torch.load(path, pickle_module=dill)
        self.net_config = checkpoint["net_config"]
        if self.net_config is not None:
            self.arch = checkpoint["net_config"]["arch"]
            if self.arch == "mlp":
                self.actor = EvolvableMLP(**checkpoint["actor_init_dict"])
                self.critic = EvolvableMLP(**checkpoint["critic_init_dict"])
            elif self.arch == "cnn":
                self.actor = EvolvableCNN(**checkpoint["actor_init_dict"])
                self.critic = EvolvableCNN(**checkpoint["critic_init_dict"])
        else:
            self.actor = MakeEvolvable(**checkpoint["actor_init_dict"])
            self.critic = MakeEvolvable(**checkpoint["critic_init_dict"])
        self.lr = checkpoint["lr"]
        self.optimizer = optim.Adam(
            [
                {"params": self.actor.parameters(), "lr": self.lr},
                {"params": self.critic.parameters(), "lr": self.lr},
            ]
        )
        self.actor.load_state_dict(checkpoint["actor_state_dict"])
        self.critic.load_state_dict(checkpoint["critic_state_dict"])
        self.optimizer.load_state_dict(checkpoint["optimizer_state_dict"])
        self.discrete_actions = checkpoint["discrete_actions"]
        self.batch_size = checkpoint["batch_size"]
        self.gamma = checkpoint["gamma"]
        self.gae_lambda = checkpoint["gae_lambda"]
        self.action_std_init = checkpoint["action_std_init"]
        self.clip_coef = checkpoint["clip_coef"]
        self.ent_coef = checkpoint["ent_coef"]
        self.vf_coef = checkpoint["vf_coef"]
        self.max_grad_norm = checkpoint["max_grad_norm"]
        self.target_kl = checkpoint["target_kl"]
        self.update_epochs = checkpoint["update_epochs"]
        self.mut = checkpoint["mutation"]
        self.index = checkpoint["index"]
        self.scores = checkpoint["scores"]
        self.fitness = checkpoint["fitness"]
        self.steps = checkpoint["steps"]



import multiprocessing as mp
import pickle as pkl
import time
from functools import partial
from typing import Generator

import redis


class Config:
    redis_host = "localhost"
    redis_port = 6379
    redis_db = 0
    client_refresh_delay = 0.01
    self_indicator = "__self__"
    init_message = "__init_message__"


def serve_class(ModelCls):
    cache_cls = pkl.dumps(ModelCls)

    class WrappedModel:
        def __init__(self, *args, **kwargs):
            self.r = redis.Redis(
                host=Config.redis_host, port=Config.redis_port, db=Config.redis_db
            )
            self.Q = initalize_server(
                self, super().__getattribute__("r"), cache_cls, args, kwargs
            )

        def __getattribute__(self, name):
            return partial(
                build_method(
                    name, super().__getattribute__("r"), super().__getattribute__("Q")
                ),
                self,
            )

        def __call__(self, *args, **kwargs):
            return build_method(
                "__call__", super().__getattribute__("r"), super().__getattribute__("Q")
            )(self, *args, **kwargs)

        def __getitem__(self, key):
            return build_method(
                "__getitem__",
                super().__getattribute__("r"),
                super().__getattribute__("Q"),
            )(self, key)

        def __setitem__(self, key, value):
            return build_method(
                "__setitem__",
                super().__getattribute__("r"),
                super().__getattribute__("Q"),
            )(self, key, value)

        def __contains__(self, key):
            return build_method(
                "__contains__",
                super().__getattribute__("r"),
                super().__getattribute__("Q"),
            )(self, key)

        def __len__(self):
            return build_method(
                "__len__", super().__getattribute__("r"), super().__getattribute__("Q")
            )(self)

    return WrappedModel


def build_method(method, r, Q):
    def call_method(self, *args, **kwargs):
        request_id = int(r.incr("request_id_counter"))
        Q.put(
            (
                request_id,
                method,
                args,
                kwargs,
            )
        )
        while not r.exists(f"result_{request_id}"):
            time.sleep(Config.client_refresh_delay)
        result = pkl.loads(r.get(f"result_{request_id}"))
        r.delete(f"result_{request_id}")
        if result == Config.self_indicator:
            return self
        return result

    return call_method


def server_process(Q, ModelCls_pkl, args, kwargs):
    r = redis.Redis(host=Config.redis_host, port=Config.redis_port, db=Config.redis_db)
    model = pkl.loads(ModelCls_pkl)(*args, **kwargs)
    while True:
        try:
            request_id, method, args, kwargs = Q.get()
            if method == Config.init_message:
                r.set(f"result_{request_id}", pkl.dumps(method))
                continue
            result = getattr(model, method)(*args, **kwargs)
            if isinstance(result, Generator):
                result = tuple(result)
            if result == model:
                result = Config.self_indicator
            r.set(f"result_{request_id}", pkl.dumps(result))
        except EOFError:
            return
        except Exception:
            raise Exception


def initalize_server(self, r, ModelCls_pkl, args, kwargs):
    Q = mp.Manager().Queue()
    p = mp.Process(target=server_process, args=(Q, ModelCls_pkl, args, kwargs))
    p.start()
    build_method(Config.init_message, r, Q)(self)
    return Q



import math
import random
from collections import defaultdict
from typing import List, Optional, Union

from tqdm.auto import tqdm

from agilerl.data.language_environment import Policy
from agilerl.utils.cache import Cache
from agilerl.wordle.wordle_env import WordleObservation
from agilerl.wordle.wordle_game import Vocabulary


class UserPolicy(Policy):
    def __init__(
        self, hint_policy: Optional[Policy], vocab: Optional[Union[str, Vocabulary]]
    ):
        super().__init__()
        self.vocab = vocab
        if isinstance(self.vocab, str):
            self.vocab = Vocabulary.from_file(self.vocab)
        self.hint_policy = hint_policy

    def act(self, obs: WordleObservation) -> str:
        print(obs.game)
        while True:
            if self.hint_policy is not None:
                want_a_hint = input("hint? ")
                if want_a_hint.lower() == "y" or want_a_hint.lower() == "yes":
                    result = self.hint_policy.act(obs)
                    print()
                    return result
            result = input("Enter a word: ")
            if len(result) != 5:
                print("Please enter a 5 letter word.")
            elif (self.vocab is not None) and (result not in self.vocab.all_vocab):
                print("Not a word. Try again.")
            else:
                break
        print()
        return result


class StartWordPolicy(Policy):
    def __init__(self, start_words: Optional[List[str]] = None):
        super().__init__()
        self.start_words = start_words
        if self.start_words is None:
            # "tales" is the optimal word under 10k_words.txt
            # "raise" is the optimal word under wordle_official.txt
            self.start_words = [
                "opera",
                "tears",
                "soare",
                "roate",
                "raise",
                "arose",
                "earls",
                "laser",
                "reals",
                "aloes",
                "reais",
                "slate",
                "sauce",
                "slice",
                "shale",
                "saute",
                "share",
                "sooty",
                "shine",
                "suite",
                "crane",
                "adieu",
                "audio",
                "stare",
                "roast",
                "ratio",
                "arise",
                "tales",
            ]

    def act(self, obs: WordleObservation) -> str:
        filtered_start_words = list(
            filter(lambda x: x in obs.game.vocab.filtered_vocab, self.start_words)
        )
        if len(filtered_start_words) == 0:
            filtered_start_words = obs.game.vocab.filtered_vocab
        return random.choice(filtered_start_words)


class OptimalPolicy(Policy):
    def __init__(
        self, start_word_policy: Optional[Policy] = None, progress_bar: bool = False
    ):
        super().__init__()
        self.start_word_policy = start_word_policy
        self.progress_bar = progress_bar
        self.cache = Cache()

    def act(self, obs: WordleObservation) -> str:
        if obs.game.state in self.cache:
            return random.choice(self.cache[obs.game.state])
        if len(obs.game.action_history) == 0 and self.start_word_policy is not None:
            return self.start_word_policy.act(obs)
        best_words = []
        best_info = float("-inf")
        for word in (
            tqdm(obs.game.vocab.filtered_vocab)
            if self.progress_bar
            else obs.game.vocab.filtered_vocab
        ):
            total_entropy = 0.0
            total = 0
            for next_state, state_count in obs.game.all_next(word):
                total_entropy += (
                    math.log(next_state.vocab.filtered_vocab_size()) * state_count
                )
                total += state_count
            info_gain = math.log(obs.game.vocab.filtered_vocab_size()) - (
                total_entropy / total
            )
            if info_gain > best_info:
                best_words, best_info = [word], info_gain
            elif info_gain == best_info:
                best_words.append(word)
        self.cache[obs.game.state] = best_words
        return random.choice(best_words)


class RepeatPolicy(Policy):
    def __init__(self, start_word_policy: Optional[Policy], first_n: Optional[int]):
        super().__init__()
        self.first_n = first_n
        self.start_word_policy = start_word_policy

    def act(self, obs: WordleObservation) -> str:
        if len(obs.game.action_history) == 0:
            if self.start_word_policy is not None:
                return self.start_word_policy.act(obs)
            return obs.game.vocab.get_random_word_all()
        if self.first_n is None:
            return random.choice(obs.game.action_history)
        return random.choice(obs.game.action_history[: self.first_n])


class RandomMixturePolicy(Policy):
    def __init__(self, prob_smart: float, vocab: Optional[Union[str, Vocabulary]]):
        super().__init__()
        self.vocab = vocab
        if isinstance(self.vocab, str):
            self.vocab = Vocabulary.from_file(self.vocab)
        self.prob_smart = prob_smart

    def act(self, obs: WordleObservation) -> str:
        if self.vocab is None:
            v = obs.game.vocab
        else:
            v = self.vocab
        if random.random() < self.prob_smart:
            if self.vocab is not None:
                v = v.update_vocab(obs.game.state)
            return v.get_random_word_filtered()
        return v.get_random_word_all()


class WrongPolicy(Policy):
    def __init__(self, vocab: Union[str, Vocabulary]):
        super().__init__()
        self.vocab = vocab
        if isinstance(self.vocab, str):
            self.vocab = Vocabulary.from_file(self.vocab)
        self.choices = set(self.vocab.all_vocab)

    def act(self, obs: WordleObservation) -> str:
        bad_options = self.choices.difference(obs.game.vocab.filtered_vocab)
        if len(bad_options) == 0:
            return self.vocab.get_random_word_all()
        return random.sample(bad_options, 1)[0]


class MixturePolicy(Policy):
    def __init__(self, prob1: float, policy1: Policy, policy2: Policy):
        super().__init__()
        self.prob1 = prob1
        self.policy1 = policy1
        self.policy2 = policy2

    def act(self, obs: WordleObservation) -> str:
        if random.random() < self.prob1:
            return self.policy1.act(obs)
        return self.policy2.act(obs)


class MonteCarloPolicy(Policy):
    def __init__(self, n_samples: int, sample_policy: Policy):
        super().__init__()
        self.n_samples = n_samples
        self.sample_policy = sample_policy

    def act(self, obs: WordleObservation) -> str:
        action_scores = defaultdict(list)
        for _ in range(self.n_samples):
            curr_obs = obs
            total_reward = 0
            while not curr_obs.game.is_terminal():
                word_choice = self.sample_policy.act(curr_obs)
                curr_obs, r, _ = curr_obs.game.next(word_choice)
                total_reward += r
            action_scores[curr_obs.action_history[len(obs.game.action_history)]].append(
                total_reward
            )
        return max(action_scores.items(), key=lambda x: sum(x[1]) / len(x[1]))[0]



import re

from agilerl.data.tokenizer import Tokenizer


class WordleTokenizer(Tokenizer):
    def __init__(self):
        self.special_vocab = [
            "<g>",
            "<b>",
            "<y>",
            "<|pad|>",
            "</a>",
            "</s>",
            "<s>",
            "<a>",
            "</eod>",
        ]
        self.vocab = list("abcdefghijklmnopqrstuvwxyz") + self.special_vocab
        self.t2i = {w: i for i, w in enumerate(self.vocab)}
        super().__init__(
            self.token_to_id("<|pad|>"),
            self.token_to_id("</s>"),
            self.token_to_id("</a>"),
            self.token_to_id("<s>"),
            self.token_to_id("<a>"),
            self.token_to_id("</eod>"),
        )

    def encode(self, str_, **kwargs):
        if isinstance(str_, str):
            special_idxs = []
            for special_char in self.special_vocab:
                special_idxs += list(
                    map(
                        lambda x: (x.start(), x.end(), self.token_to_id(special_char)),
                        re.finditer(re.escape(special_char), str_),
                    )
                )
            special_idxs.sort(key=lambda x: x[0])
            tokens = []
            curr = 0
            for s, e, tok in special_idxs:
                tokens.extend([self.token_to_id(c) for c in str_[curr:s]])
                tokens.append(tok)
                curr = e
            tokens.extend([self.token_to_id(c) for c in str_[curr:]])
            return tokens, [int(t != self.pad_token_id) for t in tokens]
        elif isinstance(str_, list):
            tokens, pads = zip(*[self.encode(item) for item in str_])
            max_len = max(map(len, tokens))
            return [
                list(item) + ([self.pad_token_id] * (max_len - len(item)))
                for item in tokens
            ], [list(item) + ([0] * (max_len - len(item))) for item in pads]
        else:
            raise ValueError("str_ must be a string or a list of strings")

    def decode(self, tokens, **kwargs):
        if len(tokens) == 0:
            return ""
        if not isinstance(tokens[0], list):
            return "".join([self.id_to_token(item) for item in tokens])
        elif isinstance(tokens[0], list):
            return [self.decode(item) for item in tokens]
        else:
            raise ValueError("tokens must be a list of ints or a list of lists of ints")

    def num_tokens(self):
        return len(self.vocab)

    def id_to_token(self, id_):
        return self.vocab[id_]

    def token_to_id(self, token):
        return self.t2i[token]

    def get_vocab(self):
        return self.vocab



import torch
import torch.nn as nn
import torch.nn.functional as F


class GumbelSoftmax(nn.Module):
    """Applies gumbel softmax function element-wise"""

    @staticmethod
    def gumbel_softmax(logits, tau=1.0, eps=1e-20):
        """Implementation of the gumbel softmax activation function

        :param logits: Tensor containing unnormalized log probabilities for each class.
        :type logits: torch.Tensor
        :param tau: Tau, defaults to 1.0
        :type tau: float, optional
        :param eps: Epsilon, defaults to 1e-20
        :type eps: float, optional
        """
        epsilon = torch.rand_like(logits)  # epsilon = U
        gumbel_noise = -torch.log(-torch.log(epsilon + eps) + eps)
        # print("GUMBEL NOISE", gumbel_noise)
        y = logits + gumbel_noise
        return F.softmax(y / tau, dim=-1)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        # print("GBSM input", input)
        return self.gumbel_softmax(input)



import random
from collections import defaultdict
from typing import Any, Dict, Optional

from agilerl.algorithms.ilql import ILQL
from agilerl.data.rl_data import DataPoint
from agilerl.wordle.wordle_dataset import WordleListDataset
from agilerl.wordle.wordle_env import WordleObservation
from agilerl.wordle.wordle_game import N_CHARS, WordleGame


class Action_Ranking_Evaluator:
    def __init__(self, branching_data: WordleListDataset) -> None:
        self.branching_data = branching_data
        self.expert_actions = defaultdict(list)
        self.non_expert_actions = defaultdict(list)
        for i in range(self.branching_data.size()):
            item = self.branching_data.get_item(i)
            assert item.meta is not None and "kind" in item.meta
            if item.meta["kind"] == "expert":
                for prefix in item.meta["prefixes"]:
                    self.expert_actions[self.hashable_state(prefix)].append(
                        item.meta["self_actions"][len(prefix[1])]
                    )
            elif item.meta["kind"] == "branch_suboptimal":
                start = item.meta["start"]
                self.non_expert_actions[self.hashable_state(start)].append(
                    item.meta["self_actions"][len(start[1])]
                )
            else:
                raise NotImplementedError
        self.states = list(
            set(self.expert_actions.keys()).intersection(
                set(self.non_expert_actions.keys())
            )
        )
        self.vocab = self.branching_data.get_item(0).meta["self"].game.vocab

    def hashable_state(self, state):
        return (
            state[0],
            tuple(state[1]),
        )

    def evaluate(self, model: ILQL, items) -> Optional[Dict[str, Any]]:
        assert not model.double_q
        tokens = model.prepare_inputs(items)["tokens"]
        total_correct = [0 for _ in range(N_CHARS + 1)]
        total_correct_target = [0 for _ in range(N_CHARS + 1)]
        total = 0
        for _ in range(tokens.shape[0]):
            s, a = random.choice(self.states)
            obs = WordleObservation(WordleGame(s, self.vocab.update_vocab(s), list(a)))
            expert_state, _, _ = obs.game.next(
                random.choice(
                    self.expert_actions[
                        self.hashable_state(
                            (
                                s,
                                a,
                            )
                        )
                    ]
                )
            )
            non_expert_state, _, _ = obs.game.next(
                random.choice(
                    self.non_expert_actions[
                        self.hashable_state(
                            (
                                s,
                                a,
                            )
                        )
                    ]
                )
            )
            expert_datapoint = DataPoint.from_obs(
                WordleObservation(expert_state),
                self.branching_data.token_reward,
                self.branching_data.tokenizer,
            )
            non_expert_datapoint = DataPoint.from_obs(
                WordleObservation(non_expert_state),
                self.branching_data.token_reward,
                self.branching_data.tokenizer,
            )
            iql_outputs = model.get_qvs([expert_datapoint, non_expert_datapoint])
            qs, target_qs, terminals = (
                iql_outputs["qs"],
                iql_outputs["target_qs"],
                iql_outputs["terminals"],
            )
            for i in range(N_CHARS + 1):
                total_correct[i] += int(
                    qs[0, (1 - terminals[0, :-1]).sum() - 1 - i]
                    > qs[1, (1 - terminals[1, :-1]).sum() - 1 - i]
                )
                total_correct_target[i] += int(
                    target_qs[0, (1 - terminals[0, :-1]).sum() - 1 - i]
                    > target_qs[1, (1 - terminals[1, :-1]).sum() - 1 - i]
                )
            total += 1
        return {
            **{
                ("q_rank_acc_-%d" % (i + 1)): (total_correct[i] / total, total)
                for i in range(N_CHARS + 1)
            },
            **{
                ("q_target_rank_acc_-%d" % (i + 1)): (
                    total_correct_target[i] / total,
                    total,
                )
                for i in range(N_CHARS + 1)
            },
        }


class Action_Ranking_Evaluator_Adversarial:
    def __init__(self, adversarial_data: WordleListDataset) -> None:
        self.adversarial_data = adversarial_data
        self.expert_actions = defaultdict(list)
        self.adversarial_actions = defaultdict(list)
        self.suboptimal_actions = defaultdict(list)
        for i in range(self.adversarial_data.size()):
            item = self.adversarial_data.get_item(i)
            assert item.meta is not None and "kind" in item.meta
            if item.meta["kind"] == "expert":
                self.expert_actions[self.hashable_state(item.meta["s_0"])].append(
                    item.meta["a_0"]
                )
                if "s_2" in item.meta:
                    self.expert_actions[self.hashable_state(item.meta["s_2"])].append(
                        item.meta["a_2"]
                    )
            elif item.meta["kind"] == "adversarial":
                self.adversarial_actions[self.hashable_state(item.meta["s_2"])].append(
                    item.meta["a_2"]
                )
            elif item.meta["kind"] == "suboptimal":
                self.suboptimal_actions[self.hashable_state(item.meta["s_0"])].append(
                    item.meta["a_0"]
                )
            else:
                raise NotImplementedError
        self.initial_states = list(
            set(self.expert_actions.keys()).intersection(
                set(self.suboptimal_actions.keys())
            )
        )
        self.branch_states = list(
            set(self.expert_actions.keys()).intersection(
                set(self.adversarial_actions.keys())
            )
        )
        self.vocab = self.adversarial_data.get_item(0).meta["self"].game.vocab

    def hashable_state(self, state):
        return (
            state[0],
            tuple(state[1]),
        )

    def evaluate(self, model: ILQL, items) -> Optional[Dict[str, Any]]:
        # evaluate Q-values for suboptimal verses expert at the first action
        # evaluate Q-values for expert versus adversarial at the third action
        assert not model.double_q
        tokens = model.prepare_inputs(items)["tokens"]
        initial_total_correct = [0 for _ in range(N_CHARS + 1)]
        initial_total_correct_target = [0 for _ in range(N_CHARS + 1)]
        branch_total_correct = [0 for _ in range(N_CHARS + 1)]
        branch_total_correct_target = [0 for _ in range(N_CHARS + 1)]
        total = 0
        for _ in range(tokens.shape[0]):
            initial_s, initial_a = random.choice(self.initial_states)
            initial_obs = WordleObservation(
                WordleGame(
                    initial_s, self.vocab.update_vocab(initial_s), list(initial_a)
                )
            )
            initial_expert_a = random.choice(
                self.expert_actions[
                    self.hashable_state(
                        (
                            initial_s,
                            initial_a,
                        )
                    )
                ]
            )
            initial_suboptimal_a = random.choice(
                self.suboptimal_actions[
                    self.hashable_state(
                        (
                            initial_s,
                            initial_a,
                        )
                    )
                ]
            )
            initial_expert_state, _, _ = initial_obs.game.next(initial_expert_a)
            initial_suboptimal_state, _, _ = initial_obs.game.next(initial_suboptimal_a)
            initial_expert_datapoint = DataPoint.from_obs(
                WordleObservation(initial_expert_state),
                self.adversarial_data.tokenizer,
                self.adversarial_data.token_reward,
            )
            initial_suboptimal_datapoint = DataPoint.from_obs(
                WordleObservation(initial_suboptimal_state),
                self.adversarial_data.tokenizer,
                self.adversarial_data.token_reward,
            )
            iql_outputs = model.get_qvs(
                [initial_expert_datapoint, initial_suboptimal_datapoint]
            )
            qs, target_qs, terminals = (
                iql_outputs["qs"],
                iql_outputs["target_qs"],
                iql_outputs["terminals"],
            )
            for i in range(N_CHARS + 1):
                initial_total_correct[i] += int(
                    qs[0, (1 - terminals[0, :-1]).sum() - 1 - i]
                    > qs[1, (1 - terminals[1, :-1]).sum() - 1 - i]
                )
                initial_total_correct_target[i] += int(
                    target_qs[0, (1 - terminals[0, :-1]).sum() - 1 - i]
                    > target_qs[1, (1 - terminals[1, :-1]).sum() - 1 - i]
                )

            branch_s, branch_a = random.choice(self.branch_states)
            branch_obs = WordleObservation(
                WordleGame(branch_s, self.vocab.update_vocab(branch_s), list(branch_a))
            )
            branch_expert_a = random.choice(
                self.expert_actions[
                    self.hashable_state(
                        (
                            branch_s,
                            branch_a,
                        )
                    )
                ]
            )
            branch_adversarial_a = random.choice(
                self.adversarial_actions[
                    self.hashable_state(
                        (
                            branch_s,
                            branch_a,
                        )
                    )
                ]
            )
            branch_expert_state, _, _ = branch_obs.game.next(branch_expert_a)
            branch_adversarial_state, _, _ = branch_obs.game.next(branch_adversarial_a)
            branch_expert_datapoint = DataPoint.from_obs(
                WordleObservation(branch_expert_state),
                self.adversarial_data.tokenizer,
                self.adversarial_data.token_reward,
            )
            branch_adversarial_datapoint = DataPoint.from_obs(
                WordleObservation(branch_adversarial_state),
                self.adversarial_data.tokenizer,
                self.adversarial_data.token_reward,
            )
            iql_outputs = model.get_qvs(
                [branch_expert_datapoint, branch_adversarial_datapoint]
            )
            qs, target_qs, terminals = (
                iql_outputs["qs"],
                iql_outputs["target_qs"],
                iql_outputs["terminals"],
            )
            for i in range(N_CHARS + 1):
                branch_total_correct[i] += int(
                    qs[0, (1 - terminals[0, :-1]).sum() - 1 - i]
                    > qs[1, (1 - terminals[1, :-1]).sum() - 1 - i]
                )
                branch_total_correct_target[i] += int(
                    target_qs[0, (1 - terminals[0, :-1]).sum() - 1 - i]
                    > target_qs[1, (1 - terminals[1, :-1]).sum() - 1 - i]
                )
            total += 1
        return {
            **{
                ("initial_q_rank_acc_-%d" % (i + 1)): (
                    initial_total_correct[i] / total,
                    total,
                )
                for i in range(N_CHARS + 1)
            },
            **{
                ("initial_q_target_rank_acc_-%d" % (i + 1)): (
                    initial_total_correct_target[i] / total,
                    total,
                )
                for i in range(N_CHARS + 1)
            },
            **{
                ("branch_q_rank_acc_-%d" % (i + 1)): (
                    branch_total_correct[i] / total,
                    total,
                )
                for i in range(N_CHARS + 1)
            },
            **{
                ("branch_q_target_rank_acc_-%d" % (i + 1)): (
                    branch_total_correct_target[i] / total,
                    total,
                )
                for i in range(N_CHARS + 1)
            },
        }



import json
import os
from collections import deque
from functools import partial

import torch
import wandb
from accelerate import Accelerator
from torch.utils.data import DataLoader
from torch.utils.data.dataset import IterableDataset
from tqdm import tqdm

from agilerl.data.rl_data import Iterable_RL_Dataset
from agilerl.data.torch_datasets import GeneralDataset, GeneralIterDataset
from agilerl.utils.ilql_utils import add_system_configs, convert_path
from agilerl.utils.load_objects import load_item
from agilerl.utils.log_utils import DistributeCombineLogs, label_logs
from agilerl.utils.torch_utils import to


def train(cfg):
    """The ILQL training function."""
    print("using config:", cfg)
    train_cfg = cfg["train"]
    train_cfg["save_checkpoint_dir"] = convert_path(train_cfg["save_checkpoint_dir"])
    train_cfg["optim_state_path"] = convert_path(train_cfg["optim_state_path"])
    wandb_cfg = cfg["wandb"]
    accelerator = Accelerator()
    system_cfg = add_system_configs(cfg, accelerator)
    print("using device:", system_cfg["device"])
    print("num processes:", system_cfg["num_processes"])
    print("using fp16:", system_cfg["use_fp16"])
    if not os.path.exists(train_cfg["save_checkpoint_dir"]):
        os.makedirs(train_cfg["save_checkpoint_dir"])
    with open(os.path.join(train_cfg["save_checkpoint_dir"], "config.json"), "w") as f:
        json.dump(cfg, f)

    if wandb_cfg["use_wandb"]:
        accelerator.wait_for_everyone()
        if accelerator.is_main_process:
            wandb.init(project=wandb_cfg["wandb_project"], config=cfg)
        accelerator.wait_for_everyone()

    raw_dataset_train = load_item(cfg["train_dataset"], system_cfg["device"])
    raw_dataset_eval = load_item(cfg["eval_dataset"], system_cfg["device"])
    if isinstance(raw_dataset_train, Iterable_RL_Dataset):
        dataset_train = GeneralIterDataset(raw_dataset_train, "cpu")
    else:
        dataset_train = GeneralDataset(raw_dataset_train, "cpu")
    if isinstance(raw_dataset_eval, Iterable_RL_Dataset):
        dataset_eval = GeneralIterDataset(raw_dataset_eval, "cpu")
    else:
        dataset_eval = GeneralDataset(raw_dataset_eval, "cpu")
    train_data_loader_kwargs = {
        "num_workers": train_cfg["dataloader_workers"],
        "batch_size": cfg["model"]["batch_size"],
        "collate_fn": dataset_train.collate,
    }
    eval_data_loader_kwargs = {
        "num_workers": train_cfg["dataloader_workers"],
        "batch_size": train_cfg["eval_bsize"],
        "collate_fn": dataset_eval.collate,
    }
    if not isinstance(dataset_train, IterableDataset):
        train_data_loader_kwargs["shuffle"] = True
    if not isinstance(dataset_eval, IterableDataset):
        eval_data_loader_kwargs["shuffle"] = True
    data_loader = DataLoader(dataset_train, **train_data_loader_kwargs)
    eval_data_loader = DataLoader(dataset_eval, **eval_data_loader_kwargs)

    evaluator = None
    if cfg["evaluator"] is not None:
        evaluator = load_item(cfg["evaluator"], system_cfg["device"])

    model = load_item(cfg["model"], system_cfg["device"])
    model.train()

    if hasattr(model, "param_groups"):
        params = [
            {
                "params": frozenset().union(*list(map(lambda x: x.parameters(), p))),
                **f(train_cfg),
            }
            for p, f in model.param_groups
        ]
        model.optimizer = torch.optim.AdamW(
            params, lr=model.lr, weight_decay=model.weight_decay
        )
    optim = model.optimizer

    if train_cfg["optim_state_path"] is not None and os.path.exists(
        train_cfg["optim_state_path"]
    ):
        print(f'loading optimizer state from: {train_cfg["optim_state_path"]}')
        optim.load_state_dict(
            torch.load(train_cfg["optim_state_path"], map_location=system_cfg["device"])
        )
        print("loaded.")
    if isinstance(dataset_train, IterableDataset) and isinstance(
        dataset_eval, IterableDataset
    ):
        model, optim = accelerator.prepare(model, optim)
    elif isinstance(dataset_train, IterableDataset):
        model, optim, eval_data_loader = accelerator.prepare(
            model, optim, eval_data_loader
        )
    elif isinstance(dataset_eval, IterableDataset):
        model, optim, data_loader = accelerator.prepare(model, optim, data_loader)
    else:
        model, optim, data_loader, eval_data_loader = accelerator.prepare(
            model, optim, data_loader, eval_data_loader
        )

    train_logs = DistributeCombineLogs(accelerator, use_wandb=wandb_cfg["use_wandb"])
    eval_logs = DistributeCombineLogs(accelerator, use_wandb=wandb_cfg["use_wandb"])
    step = 0
    best_loss = float("inf")
    saved_checkpoints = deque([])

    # bar_format = '{l_bar}{bar:10}| {n:4}/{total_fmt} [{elapsed:>7}<{remaining:>7}, {rate_fmt}{postfix}]'
    # pbar = trange(n_episodes, unit="ep", bar_format=bar_format, ascii=True)

    # RL training loop
    for epoch in range(train_cfg["epochs"]):
        for items in tqdm(data_loader, disable=not accelerator.is_local_main_process):
            items = to(items, system_cfg["device"])
            loss, logs, postproc_fs = accelerator.unwrap_model(model).get_loss(
                items, **train_cfg["loss"]
            )
            accelerator.backward(loss / train_cfg["grad_accum_steps"])
            train_logs.accum_logs(logs)
            if (step + 1) % train_cfg["grad_accum_steps"] == 0:
                optim.step()
                optim.zero_grad()
                if (
                    train_cfg["loss"]["q_loss_weight"] != 0.0
                    or train_cfg["loss"]["v_loss_weight"] != 0.0
                ):
                    accelerator.unwrap_model(model).softUpdate()
            if (train_cfg["hard_update_every"] is not None) and (
                (step + 1) % train_cfg["hard_update_every"] == 0
            ):
                accelerator.unwrap_model(model).hardUpdate()
            if (step + 1) % train_cfg["log_every"] == 0:
                train_logs.log(
                    *postproc_fs,
                    partial(label_logs, label="train"),
                    iteration=step,
                    epoch=epoch,
                )
            if (step + 1) % train_cfg["grad_accum_steps"] == 0:
                train_logs.reset_logs()
            if (step + 1) % train_cfg["eval_every"] == 0:
                model.eval()
                eval_logs.reset_logs()
                with torch.no_grad():
                    for i, eval_items in enumerate(eval_data_loader):
                        eval_items = to(eval_items, system_cfg["device"])
                        if i >= train_cfg["eval_batches"]:
                            break
                        _, logs, postproc_fs = accelerator.unwrap_model(model).get_loss(
                            eval_items, **train_cfg["loss"]
                        )
                        if evaluator is not None:
                            evaluator_logs = evaluator.evaluate(
                                accelerator.unwrap_model(model), eval_items
                            )
                            if evaluator_logs is not None:
                                logs["evaluation"] = evaluator_logs
                        eval_logs.accum_logs(logs)
                eval_label = "eval"
                eval_total_logs = eval_logs.log(
                    *postproc_fs,
                    partial(label_logs, label=eval_label),
                    iteration=step,
                    epoch=epoch,
                )
                accelerator.wait_for_everyone()
                if accelerator.is_main_process:
                    if eval_total_logs[eval_label]["loss"] < best_loss:
                        print("new best eval loss! Saving ...")
                        if not os.path.exists(train_cfg["save_checkpoint_dir"]):
                            os.makedirs(train_cfg["save_checkpoint_dir"])
                        torch.save(
                            accelerator.unwrap_model(model).state_dict(),
                            os.path.join(train_cfg["save_checkpoint_dir"], "model.pkl"),
                        )
                        torch.save(
                            optim.state_dict(),
                            os.path.join(train_cfg["save_checkpoint_dir"], "optim.pkl"),
                        )
                        print("saved.")
                        best_loss = eval_total_logs[eval_label]["loss"]
                accelerator.wait_for_everyone()
                model.train()
            if (
                train_cfg["save_every"] is not None
                and (step + 1) % train_cfg["save_every"] == 0
            ):
                accelerator.wait_for_everyone()
                if accelerator.is_main_process:
                    print("saving checkpoint...")
                    if not os.path.exists(train_cfg["save_checkpoint_dir"]):
                        os.makedirs(train_cfg["save_checkpoint_dir"])
                    if (train_cfg["max_checkpoints"] is not None) and (
                        len(saved_checkpoints) >= train_cfg["max_checkpoints"]
                    ):
                        os.system("rm -rf %s" % (saved_checkpoints.popleft()))
                    torch.save(
                        accelerator.unwrap_model(model).state_dict(),
                        os.path.join(
                            train_cfg["save_checkpoint_dir"], "model_%d.pkl" % (step)
                        ),
                    )
                    saved_checkpoints.append(
                        os.path.join(
                            train_cfg["save_checkpoint_dir"], "model_%d.pkl" % (step)
                        )
                    )
                    print("saved.")
                accelerator.wait_for_everyone()
            step += 1
            if train_cfg["max_steps"] is not None and step >= train_cfg["max_steps"]:
                return



import json

import torch

from agilerl.algorithms.bc_lm import BC_LM, BC_Evaluator, BC_Policy
from agilerl.algorithms.ilql import (
    ILQL,
    ILQL_Evaluator,
    ILQL_Policy,
    TopAdvantageNGrams,
)
from agilerl.data.rl_data import ConstantTokenReward, SepcifiedTokenReward
from agilerl.networks.evolvable_gpt import EvolvableGPT
from agilerl.utils.ilql_utils import convert_path
from agilerl.wordle.policy import (
    MixturePolicy,
    MonteCarloPolicy,
    OptimalPolicy,
    RandomMixturePolicy,
    RepeatPolicy,
    StartWordPolicy,
    UserPolicy,
    WrongPolicy,
)
from agilerl.wordle.wordle_dataset import (
    WordleHumanDataset,
    WordleIterableDataset,
    WordleListDataset,
)
from agilerl.wordle.wordle_env import WordleEnvironment
from agilerl.wordle.wordle_evaluators import (
    Action_Ranking_Evaluator,
    Action_Ranking_Evaluator_Adversarial,
)
from agilerl.wordle.wordle_game import Vocabulary

registry = {}
cache = {}


def register(name):
    def add_f(f):
        registry[name] = f
        return f

    return add_f


def load_item(config, *args, verbose=True):
    config = config.copy()
    name = config.pop("name")
    if name not in registry:
        raise NotImplementedError
    if "cache_id" in config:
        if (name, config["cache_id"]) in cache:
            if verbose:
                print(f'loading from cache ({name}, {config["cache_id"]})')
            return cache[(name, config["cache_id"])]
    if verbose:
        print(f"loading {name}: {config}")
    item = registry[name](config, *args, verbose=verbose)
    if "cache_id" in config:
        print(f'saving to cache ({name}, {config["cache_id"]})')
        cache[(name, config["cache_id"])] = item
    return item


def load_model(config, model, device, verbose=True):
    model = model.to(device)
    if config["checkpoint_path"] is not None:
        if verbose:
            print(
                "loading %s state dict from: %s"
                % (config["name"], convert_path(config["checkpoint_path"]))
            )
        chkpt_state_dict = torch.load(
            convert_path(config["checkpoint_path"]), map_location="cpu"
        )
        model.load_state_dict(chkpt_state_dict, strict=config["strict_load"])
        if verbose:
            print("loaded.")
    elif config["gpt_pretrained"]:
        if verbose:
            print(
                "loading %s state dict from: %s"
                % (config["name"], convert_path(config["gpt_checkpoint_path"]))
            )
        pretrained_state_dict = convert_path(config["gpt_checkpoint_path"])
        model.model = EvolvableGPT.from_pretrained(
            model_type=config["gpt_model_type"],
            override_args=model.net_config,
            custom_sd=pretrained_state_dict,
        )
        model.copy_model_to_actor_target()
    return model


@register("constant_token_reward")
def load_constant_token_reward(config, device, verbose=True):
    return ConstantTokenReward(config["c"])


@register("specified_token_reward")
def load_specified_token_reward(config, device, verbose=True):
    with open(convert_path(config["token_file"])) as f:
        token_data = {int(k): v for k, v in json.load(f).items()}
    return SepcifiedTokenReward(token_data, config["scale"], config["shift"])


@register("gpt2")
def load_gpt2(config, verbose=True):
    if config["from_pretrained"]:
        return EvolvableGPT.from_pretrained(config["gpt2_type"])
    return EvolvableGPT()


@register("bc_lm")
def load_bc_lm(config, device, verbose=True):
    dataset = load_item(config["dataset"], device, verbose=verbose)
    config.pop("dataset")
    load_config = config.pop("load")
    model = BC_LM(**config, dataset=dataset, device=device)
    return load_model(load_config, model, device, verbose=verbose)


@register("bc_policy")
def load_bc_policy(config, device, verbose=True):
    bc_lm = load_item(config["bc_lm"], device, verbose=verbose)
    return BC_Policy(bc_lm, config["kind"], **config["generation_kwargs"])


@register("bc_evaluator")
def load_bc_evaluator(config, device, verbose=True):
    env = load_item(config["env"], device, verbose=verbose)
    return BC_Evaluator(
        env, config["env"], config["kind"], **config["generation_kwargs"]
    )


@register("per_token_iql")
def load_per_token_iql(config, device, verbose=True):
    dataset = load_item(config["dataset"], device, verbose=verbose)
    config.pop("dataset")
    load_config = config.pop("load")
    model = ILQL(**config, dataset=dataset, device=device)
    return load_model(load_config, model, device, verbose=verbose)


@register("iql_policy")
def load_iql_policy(config, device, verbose=True):
    iql_model = load_item(config["iql_model"], device, verbose=verbose)
    return ILQL_Policy(iql_model, config["kind"], **config["generation_kwargs"])


@register("iql_evaluator")
def load_iql_evaluator(config, device, verbose=True):
    env = load_item(config["env"], device, verbose=verbose)
    return ILQL_Evaluator(
        env, config["verbose"], config["kind"], **config["generation_kwargs"]
    )


@register("top_advantage_n_grams")
def load_top_advantage_n_grams(config, device, verbose=True):
    data = load_item(config["data"], device, verbose=verbose)
    return TopAdvantageNGrams(
        data, config["print_every"], config["print_k"], config["n_gram"]
    )


@register("vocab")
def load_vocab(config, verbose=True):
    vocab = Vocabulary.from_file(
        convert_path(config["vocab_path"]), config["fill_cache"]
    )
    if config["cache_path"] is not None:
        if verbose:
            print("loading vocab cache from: %s" % convert_path(config["cache_path"]))
        vocab.cache.load(convert_path(config["cache_path"]))
        if verbose:
            print("loaded.")
    return vocab


@register("wordle_env")
def load_wordle_environment(config, device, verbose=True):
    vocab = load_item(config["vocab"], verbose=verbose)
    return WordleEnvironment(vocab)


@register("user_policy")
def load_user_policy(config, device, verbose=True):
    vocab, hint_policy = None, None
    if config["hint_policy"] is not None:
        hint_policy = load_item(config["hint_policy"], device, verbose=verbose)
    if config["vocab"] is not None:
        vocab = load_item(config["vocab"], verbose=verbose)
    return UserPolicy(hint_policy=hint_policy, vocab=vocab)


@register("start_word_policy")
def load_start_word_policy(config, device, verbose=True):
    return StartWordPolicy(config["start_words"])


@register("optimal_policy")
def load_optimal_policy(config, device, verbose=True):
    start_word_policy = None
    if config["start_word_policy"] is not None:
        start_word_policy = load_item(
            config["start_word_policy"], device, verbose=verbose
        )
    policy = OptimalPolicy(
        start_word_policy=start_word_policy, progress_bar=config["progress_bar"]
    )
    if config["cache_path"] is not None:
        if verbose:
            print(
                "loading optimal policy cache from: %s"
                % convert_path(config["cache_path"])
            )
        policy.cache.load(convert_path(config["cache_path"]))
        if verbose:
            print("loaded.")
    return policy


@register("wrong_policy")
def load_wrong_policy(config, device, verbose=True):
    vocab = load_item(config["vocab"], verbose=verbose)
    return WrongPolicy(vocab=vocab)


@register("repeat_policy")
def load_repeat_policy(config, device, verbose=True):
    start_word_policy = None
    if config["start_word_policy"] is not None:
        start_word_policy = load_item(
            config["start_word_policy"], device, verbose=verbose
        )
    return RepeatPolicy(start_word_policy=start_word_policy, first_n=config["first_n"])


@register("mixture_policy")
def load_mixture_policy(config, device, verbose=True):
    policy1 = load_item(config["policy1"], device, verbose=verbose)
    policy2 = load_item(config["policy2"], device, verbose=verbose)
    return MixturePolicy(config["prob1"], policy1, policy2)


@register("random_mixture_policy")
def load_random_mixture_policy(config, device, verbose=True):
    vocab = None
    if config["vocab"] is not None:
        vocab = load_item(config["vocab"], verbose=verbose)
    return RandomMixturePolicy(prob_smart=config["prob_smart"], vocab=vocab)


@register("monte_carlo_policy")
def load_monte_carlo_policy(config, device, verbose=True):
    sample_policy = load_item(config["sample_policy"], device, verbose=verbose)
    return MonteCarloPolicy(n_samples=config["n_samples"], sample_policy=sample_policy)


@register("wordle_iterable_dataset")
def load_wordle_iterable_dataset(config, device, verbose=True):
    policy = load_item(config["policy"], device, verbose=verbose)
    vocab = load_item(config["vocab"], verbose=verbose)
    token_reward = load_item(config["token_reward"], device, verbose=verbose)
    return WordleIterableDataset(
        policy, vocab, max_len=config["max_len"], token_reward=token_reward
    )


@register("wordle_dataset")
def load_wordle_dataset(config, device, verbose=True):
    if config["vocab"] is not None:
        vocab = load_item(config["vocab"], verbose=verbose)
    else:
        vocab = None
    token_reward = load_item(config["token_reward"], device, verbose=verbose)
    return WordleListDataset.from_file(
        convert_path(config["file_path"]), config["max_len"], vocab, token_reward
    )


@register("wordle_human_dataset")
def load_human_dataset(config, device, verbose=True):
    token_reward = load_item(config["token_reward"], device, verbose=verbose)
    game_indexes = None
    if config["index_file"] is not None:
        with open(convert_path(config["index_file"])) as f:
            game_indexes = json.load(f)
    return WordleHumanDataset.from_file(
        convert_path(config["file_path"]),
        config["use_true_word"],
        config["max_len"],
        token_reward,
        game_indexes,
        config["top_p"],
    )


@register("action_ranking_evaluator")
def load_action_ranking_evaluator(config, device, verbose=True):
    branching_data = load_item(config["branching_data"], device, verbose=verbose)
    return Action_Ranking_Evaluator(branching_data)


@register("action_ranking_evaluator_adversarial")
def load_action_ranking_evaluator_adversarial(config, device, verbose=True):
    adversarial_data = load_item(config["adversarial_data"], device, verbose=verbose)
    return Action_Ranking_Evaluator_Adversarial(adversarial_data)



from typing import List, Optional, Tuple

from agilerl.data.language_environment import Language_Environment, Language_Observation
from agilerl.wordle.wordle_game import Vocabulary, WordleGame


class WordleObservation(Language_Observation):
    def __init__(self, game: WordleGame):
        self.game = game

    def to_sequence(self) -> Tuple[List[Tuple[str, Optional[float]]], bool]:
        return self.game.transition_sequence()

    def __str__(self) -> str:
        return str(self.game)


class WordleEnvironment(Language_Environment):
    def __init__(self, vocab: Vocabulary):
        self.vocab = vocab
        self.state = self.reset()

    def step(self, action: str) -> Tuple[WordleObservation, float, bool]:
        wordle_game, r, t = self.state.game.next(action)
        self.state = WordleObservation(wordle_game)
        return self.state, r, t

    def reset(self) -> WordleObservation:
        self.state = WordleObservation(WordleGame.initialize(self.vocab))
        return self.state

    def is_terminal(self) -> bool:
        return self.state.game.is_terminal()



import os
import pickle as pkl
import sys
from collections import defaultdict
from typing import Any, Dict, List

import torch


def convert_path(path):
    if path is None:
        return None
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), "../../", path)


def add_system_configs(cfg, accelerator):
    cfg["system"] = {}
    cfg["system"]["device"] = str(accelerator.device)
    cfg["system"]["num_processes"] = accelerator.num_processes
    cfg["system"]["use_fp16"] = accelerator.use_fp16
    return cfg["system"]


def to_bin(n, pad_to_size=None):
    bins = to_bin(n // 2) + [n % 2] if n > 1 else [n]
    if pad_to_size is None:
        return bins
    return ([0] * (pad_to_size - len(bins))) + bins


def strip_from_end(str_item, strip_key):
    return strip_from_beginning(str_item[::-1], strip_key[::-1])[::-1]


def strip_from_beginning(str_item, strip_key):
    if str_item[: len(strip_key)] == strip_key:
        return str_item[len(strip_key) :]
    return str_item


def migrate_pkl(pkl_path):
    from wordle import wordle_game

    sys.modules["wordle.wordle_mdp"] = wordle_game
    with open(pkl_path, "rb") as f:
        d = pkl.load(f)
    del sys.modules["wordle.wordle_mdp"]
    with open(pkl_path, "wb") as f:
        pkl.dump(d, f)


def convert_old_ad_model(model_path, dataset, multiple_transformers=False):
    w = torch.load(model_path, map_location="cpu")
    embs = w["model.transformer.wte.weight"]
    if embs.shape[0] < dataset.tokenizer.num_tokens():
        w["model.transformer.wte.weight"] = torch.cat(
            (
                embs,
                torch.zeros(
                    dataset.tokenizer.num_tokens() - embs.shape[0], embs.shape[1]
                ),
            ),
            dim=0,
        )
    head = w["model.lm_head.weight"]
    if head.shape[0] < dataset.tokenizer.num_tokens():
        w["model.lm_head.weight"] = torch.cat(
            (
                head,
                torch.zeros(
                    dataset.tokenizer.num_tokens() - head.shape[0], head.shape[1]
                ),
            ),
            dim=0,
        )
    if multiple_transformers:
        for k, v in list(w.items()):
            if k.startswith("model."):
                w["lm_policy." + k[len("model.") :]] = v.clone()
                w["lm_target." + k[len("model.") :]] = v.clone()
    torch.save(w, model_path)


def stack_dicts(dicts: List[Dict[str, Any]]):
    stacked = defaultdict(list)
    for item in dicts:
        for k, v in item.items():
            stacked[k].append(v)
    # rough quick check for dicts have the same set of keys, could be more rigorous
    assert len(set(map(len, stacked.values()))) == 1
    return stacked


def unstack_dicts(stacked_dict: Dict[str, List[Any]]):
    dict_len = set(map(lambda x: len(list(x)), stacked_dict.values()))
    assert len(dict_len) == 1
    dicts = [{} for _ in range(dict_len.pop())]
    for k, v in stacked_dict.items():
        for i, item in enumerate(v):
            dicts[i][k] = item
    return dicts


class PrecisionRecallAcc:
    def __init__(self, classes) -> None:
        self.precisions = {class_: [0, 0] for class_ in classes}
        self.recalls = {class_: [0, 0] for class_ in classes}
        self.correct = 0
        self.total = 0

    def add_item(self, prediction_type: Any, actual_type: Any, correct: bool):
        correct = int(correct)
        if prediction_type in self.precisions:
            self.precisions[prediction_type][0] += correct
            self.precisions[prediction_type][1] += 1
        if actual_type in self.recalls:
            self.recalls[actual_type][0] += correct
            self.recalls[actual_type][1] += 1
        self.correct += correct
        self.total += 1

    def return_summary(self):
        logs = {}
        logs["accuracy"] = (self.correct / self.total, self.total)
        for k, (a, b) in self.precisions.items():
            logs[str(k) + "_precision"] = (a / b if b != 0 else -1, b)
        for k, (a, b) in self.recalls.items():
            logs[str(k) + "_recall"] = (a / b if b != 0 else -1, b)
        return logs



import copy
import random
import warnings

import dill
import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim

from agilerl.networks.evolvable_cnn import EvolvableCNN
from agilerl.networks.evolvable_mlp import EvolvableMLP
from agilerl.wrappers.make_evolvable import MakeEvolvable


class DDPG:
    """The DDPG algorithm class. DDPG paper: https://arxiv.org/abs/1509.02971

    :param state_dim: State observation dimension
    :type state_dim: list[int]
    :param action_dim: Action dimension
    :type action_dim: int
    :param one_hot: One-hot encoding, used with discrete observation spaces
    :type one_hot: bool
    :param max_action: Upper bound of the action space, defaults to 1
    :type max_action: float, optional
    :param min_action: Lower bound of the action space, defaults to -1
    :type min_action: float, optional
    :param index: Index to keep track of object instance during tournament selection and mutation, defaults to 0
    :type index: int, optional
    :param net_config: Network configuration, defaults to mlp with hidden size [64,64]
    :type net_config: dict, optional
    :param batch_size: Size of batched sample from replay buffer for learning, defaults to 64
    :type batch_size: int, optional
    :param lr: Learning rate for optimizer, defaults to 1e-4
    :type lr: float, optional
    :param learn_step: Learning frequency, defaults to 5
    :type learn_step: int, optional
    :param gamma: Discount factor, defaults to 0.99
    :type gamma: float, optional
    :param tau: For soft update of target network parameters, defaults to 1e-3
    :type tau: float, optional
    :param mutation: Most recent mutation to agent, defaults to None
    :type mutation: str, optional
    :param policy_freq: Frequency of critic network updates compared to policy network, defaults to 2
    :type policy_freq: int, optional
    :param actor_network: Custom actor network, defaults to None
    :type actor_network: nn.Module, optional
    :param critic_network: Custom critic network, defaults to None
    :type critic_network: nn.Module, optional
    :param device: Device for accelerated computing, 'cpu' or 'cuda', defaults to 'cpu'
    :type device: str, optional
    :param accelerator: Accelerator for distributed computing, defaults to None
    :type accelerator: accelerate.Accelerator(), optional
    :param wrap: Wrap models for distributed training upon creation, defaults to True
    :type wrap: bool, optional
    """

    def __init__(
        self,
        state_dim,
        action_dim,
        one_hot,
        max_action=1,
        min_action=-1,
        index=0,
        net_config={"arch": "mlp", "h_size": [64, 64]},
        batch_size=64,
        lr=1e-4,
        learn_step=5,
        gamma=0.99,
        tau=1e-3,
        mutation=None,
        policy_freq=2,
        actor_network=None,
        critic_network=None,
        device="cpu",
        accelerator=None,
        wrap=True,
    ):
        assert isinstance(
            state_dim, (list, tuple)
        ), "State dimension must be a list or tuple."
        assert isinstance(
            action_dim, (int, np.integer)
        ), "Action dimension must be an integer."
        assert isinstance(
            one_hot, bool
        ), "One-hot encoding flag must be boolean value True or False."
        assert isinstance(
            max_action, (float, int, np.floating, np.integer)
        ), "Max action must be a float or integer."
        assert isinstance(
            min_action, (float, int, np.floating, np.integer)
        ), "Min action must be a float or integer."
        assert max_action > min_action, "Max action must be greater than min action."
        assert max_action > 0, "Max action must be greater than zero."
        assert min_action <= 0, "Min action must be less than or equal to zero."
        assert isinstance(index, int), "Agent index must be an integer."
        assert isinstance(batch_size, int), "Batch size must be an integer."
        assert batch_size >= 1, "Batch size must be greater than or equal to one."
        assert isinstance(lr, float), "Learning rate must be a float."
        assert lr > 0, "Learning rate must be greater than zero."
        assert isinstance(learn_step, int), "Learn step rate must be an integer."
        assert learn_step >= 1, "Learn step must be greater than or equal to one."
        assert isinstance(gamma, (float, int)), "Gamma must be a float."
        assert isinstance(tau, float), "Tau must be a float."
        assert tau > 0, "Tau must be greater than zero."
        assert isinstance(policy_freq, int), "Policy frequency must be an integer."
        assert (
            policy_freq >= 1
        ), "Policy frequency must be greater than or equal to one."
        assert (
            isinstance(actor_network, nn.Module) or actor_network is None
        ), "Actor network must be an nn.Module or None."
        assert (
            isinstance(critic_network, nn.Module) or critic_network is None
        ), "Critic network must be an nn.Module or None."
        if (actor_network is not None) != (critic_network is not None):  # XOR operation
            warnings.warn(
                "Actor and critic networks must both be supplied to use custom networks. Defaulting to net config."
            )
        assert isinstance(
            wrap, bool
        ), "Wrap models flag must be boolean value True or False."

        self.algo = "DDPG"
        self.state_dim = state_dim
        self.action_dim = action_dim
        self.one_hot = one_hot
        self.max_action = max_action
        self.min_action = min_action
        self.net_config = net_config
        self.batch_size = batch_size
        self.lr = lr
        self.learn_step = learn_step
        self.gamma = gamma
        self.tau = tau
        self.mut = mutation
        self.policy_freq = policy_freq
        self.actor_network = actor_network
        self.critic_network = critic_network
        self.device = device
        self.accelerator = accelerator

        self.index = index
        self.scores = []
        self.fitness = []
        self.steps = [0]

        if self.actor_network is not None and self.critic_network is not None:
            self.actor = actor_network
            self.critic = critic_network
            self.net_config = None
        else:
            # model
            assert isinstance(self.net_config, dict), "Net config must be a dictionary."
            assert (
                "arch" in self.net_config.keys()
            ), "Net config must contain arch: 'mlp' or 'cnn'."
            if self.min_action < 0:
                output_activation = "Tanh"
            else:
                output_activation = "Sigmoid"
            if self.net_config["arch"] == "mlp":  # Multi-layer Perceptron
                assert (
                    "h_size" in self.net_config.keys()
                ), "Net config must contain h_size: int."
                assert isinstance(
                    self.net_config["h_size"], list
                ), "Net config h_size must be a list."
                assert (
                    len(self.net_config["h_size"]) > 0
                ), "Net config h_size must contain at least one element."
                self.actor = EvolvableMLP(
                    num_inputs=state_dim[0],
                    num_outputs=action_dim,
                    hidden_size=self.net_config["h_size"],
                    mlp_output_activation=output_activation,
                    device=self.device,
                    accelerator=self.accelerator,
                )
                self.critic = EvolvableMLP(
                    num_inputs=state_dim[0] + action_dim,
                    num_outputs=1,
                    hidden_size=self.net_config["h_size"],
                    device=self.device,
                    accelerator=self.accelerator,
                )
            elif self.net_config["arch"] == "cnn":  # Convolutional Neural Network
                for key in ["c_size", "k_size", "s_size", "h_size"]:
                    assert (
                        key in self.net_config.keys()
                    ), f"Net config must contain {key}: int."
                    assert isinstance(
                        self.net_config[key], list
                    ), f"Net config {key} must be a list."
                    assert (
                        len(self.net_config[key]) > 0
                    ), f"Net config {key} must contain at least one element."
                assert (
                    "normalize" in self.net_config.keys()
                ), "Net config must contain normalize: True or False."
                assert isinstance(
                    self.net_config["normalize"], bool
                ), "Net config normalize must be boolean value True or False."
                self.actor = EvolvableCNN(
                    input_shape=state_dim,
                    num_actions=action_dim,
                    channel_size=self.net_config["c_size"],
                    kernel_size=self.net_config["k_size"],
                    stride_size=self.net_config["s_size"],
                    hidden_size=self.net_config["h_size"],
                    normalize=self.net_config["normalize"],
                    mlp_activation="Tanh",
                    mlp_output_activation=output_activation,
                    device=self.device,
                    accelerator=self.accelerator,
                )
                self.critic = EvolvableCNN(
                    input_shape=state_dim,
                    num_actions=action_dim,
                    channel_size=self.net_config["c_size"],
                    kernel_size=self.net_config["k_size"],
                    stride_size=self.net_config["s_size"],
                    hidden_size=self.net_config["h_size"],
                    normalize=self.net_config["normalize"],
                    mlp_activation="Tanh",
                    mlp_output_activation=None,
                    critic=True,
                    device=self.device,
                    accelerator=self.accelerator,
                )

        self.actor_target = copy.deepcopy(self.actor)
        self.critic_target = copy.deepcopy(self.critic)
        self.actor_target.load_state_dict(self.actor.state_dict())
        self.critic_target.load_state_dict(self.critic.state_dict())
        self.actor_optimizer_type = optim.Adam(self.actor.parameters(), lr=self.lr)
        self.critic_optimizer_type = optim.Adam(self.critic.parameters(), lr=self.lr)

        self.arch = (
            self.net_config["arch"] if self.net_config is not None else self.actor.arch
        )

        if self.accelerator is not None:
            self.actor_optimizer = self.actor_optimizer_type
            self.critic_optimizer = self.critic_optimizer_type
            if wrap:
                self.wrap_models()
        else:
            self.actor = self.actor.to(self.device)
            self.actor_target = self.actor_target.to(self.device)
            self.critic = self.critic.to(self.device)
            self.critic_target = self.critic_target.to(self.device)
            self.actor_optimizer = self.actor_optimizer_type
            self.critic_optimizer = self.critic_optimizer_type

        self.criterion = nn.MSELoss()

    def scale_to_action_space(self, action):
        """Scales actions to action space defined by self.min_action and self.max_action.

        :param action: Action to be scaled
        :type action: numpy.ndarray
        """
        return np.where(action > 0, action * self.max_action, action * -self.min_action)

    def getAction(self, state, epsilon=0):
        """Returns the next action to take in the environment.
        Epsilon is the probability of taking a random action, used for exploration.
        For epsilon-greedy behaviour, set epsilon to 0.

        :param state: Environment observation, or multiple observations in a batch
        :type state: float or list[float]
        :param epsilon: Probablilty of taking a random action for exploration, defaults to 0
        :type epsilon: float, optional
        """
        state = torch.from_numpy(state).float()
        if self.accelerator is None:
            state = state.to(self.device)
        else:
            state = state.to(self.accelerator.device)

        if self.one_hot:
            state = (
                nn.functional.one_hot(state.long(), num_classes=self.state_dim[0])
                .float()
                .squeeze()
            )

        if len(state.size()) < 2:
            state = state.unsqueeze(0)

        # epsilon-greedy
        if random.random() < epsilon:
            action = (
                (self.max_action - self.min_action)
                * np.random.rand(state.size()[0], self.action_dim).astype("float32")
            ) + self.min_action
        else:
            self.actor.eval()
            with torch.no_grad():
                action_values = self.actor(state)
            self.actor.train()

            action = self.scale_to_action_space(action_values.cpu().data.numpy())

        return action

    def learn(self, experiences, noise_clip=0.5, policy_noise=0.2):
        """Updates agent network parameters to learn from experiences.

        :param experience: List of batched states, actions, rewards, next_states, dones in that order.
        :type experience: list[torch.Tensor[float]]
        :param noise_clip: Maximum noise limit to apply to actions, defaults to 0.5
        :type noise_clip: float, optional
        :param policy_noise: Standard deviation of noise applied to policy, defaults to 0.2
        :type policy_noise: float, optional
        """
        states, actions, rewards, next_states, dones = experiences
        if self.accelerator is not None:
            states = states.to(self.accelerator.device)
            actions = actions.to(self.accelerator.device)
            rewards = rewards.to(self.accelerator.device)
            next_states = next_states.to(self.accelerator.device)
            dones = dones.to(self.accelerator.device)

        if self.one_hot:
            states = (
                nn.functional.one_hot(states.long(), num_classes=self.state_dim[0])
                .float()
                .squeeze()
            )
            next_states = (
                nn.functional.one_hot(next_states.long(), num_classes=self.state_dim[0])
                .float()
                .squeeze()
            )

        if self.arch == "mlp":
            input_combined = torch.cat([states, actions], 1)
            q_value = self.critic(input_combined)
        elif self.arch == "cnn":
            q_value = self.critic(states, actions)

        next_actions = self.actor_target(next_states)
        # Scale actions
        next_actions = torch.where(
            next_actions > 0,
            next_actions * self.max_action,
            next_actions * -self.min_action,
        )
        noise = actions.data.normal_(0, policy_noise)
        noise = noise.clamp(-noise_clip, noise_clip)
        next_actions = next_actions + noise

        if self.arch == "mlp":
            next_input_combined = torch.cat([next_states, next_actions], 1)
            q_value_next_state = self.critic_target(next_input_combined)
        elif self.arch == "cnn":
            q_value_next_state = self.critic_target(next_states, next_actions)

        y_j = rewards + ((1 - dones) * self.gamma * q_value_next_state).detach()

        critic_loss = self.criterion(q_value, y_j)

        # critic loss backprop
        self.critic_optimizer.zero_grad()
        if self.accelerator is not None:
            self.accelerator.backward(critic_loss)
        else:
            critic_loss.backward()
        self.critic_optimizer.step()

        # update actor and targets every policy_freq episodes
        if len(self.scores) % self.policy_freq == 0:
            policy_actions = self.actor.forward(states)
            policy_actions = torch.where(
                policy_actions > 0,
                policy_actions * self.max_action,
                policy_actions * -self.min_action,
            )
            if self.arch == "mlp":
                input_combined = torch.cat([states, policy_actions], 1)
                actor_loss = -self.critic(input_combined).mean()
            elif self.arch == "cnn":
                actor_loss = -self.critic(states, policy_actions).mean()

            # actor loss backprop
            self.actor_optimizer.zero_grad()
            if self.accelerator is not None:
                self.accelerator.backward(actor_loss)
            else:
                actor_loss.backward()
            self.actor_optimizer.step()

            self.softUpdate(self.actor, self.actor_target)
            self.softUpdate(self.critic, self.critic_target)

            return actor_loss.item(), critic_loss.item()
        else:
            return None, critic_loss.item()

    def softUpdate(self, net, target):
        """Soft updates target network."""
        for eval_param, target_param in zip(net.parameters(), target.parameters()):
            target_param.data.copy_(
                self.tau * eval_param.data + (1.0 - self.tau) * target_param.data
            )

    def test(self, env, swap_channels=False, max_steps=500, loop=3):
        """Returns mean test score of agent in environment with epsilon-greedy policy.

        :param env: The environment to be tested in
        :type env: Gym-style environment
        :param swap_channels: Swap image channels dimension from last to first [H, W, C] -> [C, H, W], defaults to False
        :type swap_channels: bool, optional
        :param max_steps: Maximum number of testing steps, defaults to 500
        :type max_steps: int, optional
        :param loop: Number of testing loops/episodes to complete. The returned score is the mean. Defaults to 3
        :type loop: int, optional
        """
        with torch.no_grad():
            rewards = []
            for i in range(loop):
                state = env.reset()[0]
                score = 0
                for idx_step in range(max_steps):
                    if swap_channels:
                        if not hasattr(env, "num_envs"):
                            state = np.expand_dims(state, 0)
                        state = np.moveaxis(state, [3], [1])
                    action = self.getAction(state, epsilon=0)
                    if not hasattr(env, "num_envs"):
                        action = action[0]
                    state, reward, done, trunc, _ = env.step(action)
                    if hasattr(env, "num_envs"):
                        done = done[0]
                        trunc = trunc[0]
                        reward = reward[0]
                    score += reward
                    if done or trunc:
                        break
                rewards.append(score)
        mean_fit = np.mean(rewards)
        self.fitness.append(mean_fit)
        return mean_fit

    def clone(self, index=None, wrap=True):
        """Returns cloned agent identical to self.

        :param index: Index to keep track of agent for tournament selection and mutation, defaults to None
        :type index: int, optional
        """
        if index is None:
            index = self.index

        clone = type(self)(
            state_dim=self.state_dim,
            action_dim=self.action_dim,
            one_hot=self.one_hot,
            index=index,
            net_config=self.net_config,
            actor_network=self.actor_network,
            critic_network=self.critic_network,
            batch_size=self.batch_size,
            lr=self.lr,
            learn_step=self.learn_step,
            gamma=self.gamma,
            tau=self.tau,
            mutation=self.mut,
            policy_freq=self.policy_freq,
            device=self.device,
            accelerator=self.accelerator,
            wrap=wrap,
        )

        if self.accelerator is not None:
            self.unwrap_models()
        actor = self.actor.clone()
        actor_target = self.actor_target.clone()
        critic = self.critic.clone()
        critic_target = self.critic_target.clone()
        actor_optimizer = optim.Adam(actor.parameters(), lr=clone.lr)
        critic_optimizer = optim.Adam(critic.parameters(), lr=clone.lr)
        clone.actor_optimizer_type = actor_optimizer
        clone.critic_optimizer_type = critic_optimizer

        if self.accelerator is not None:
            if wrap:
                (
                    clone.actor,
                    clone.actor_target,
                    clone.critic,
                    clone.critic_target,
                    clone.actor_optimizer,
                    clone.critic_optimizer,
                ) = self.accelerator.prepare(
                    actor,
                    actor_target,
                    critic,
                    critic_target,
                    actor_optimizer,
                    critic_optimizer,
                )
            else:
                (
                    clone.actor,
                    clone.actor_target,
                    clone.critic,
                    clone.critic_target,
                    clone.actor_optimizer,
                    clone.critic_optimizer,
                ) = (
                    actor,
                    actor_target,
                    critic,
                    critic_target,
                    actor_optimizer,
                    critic_optimizer,
                )
        else:
            clone.actor = actor.to(self.device)
            clone.actor_target = actor_target.to(self.device)
            clone.critic = critic.to(self.device)
            clone.critic_target = critic_target.to(self.device)
            clone.actor_optimizer = actor_optimizer
            clone.critic_optimizer = critic_optimizer

        clone.fitness = copy.deepcopy(self.fitness)
        clone.steps = copy.deepcopy(self.steps)
        clone.scores = copy.deepcopy(self.scores)

        return clone

    def wrap_models(self):
        if self.accelerator is not None:
            (
                self.actor,
                self.actor_target,
                self.critic,
                self.critic_target,
                self.actor_optimizer,
                self.critic_optimizer,
            ) = self.accelerator.prepare(
                self.actor,
                self.actor_target,
                self.critic,
                self.critic_target,
                self.actor_optimizer_type,
                self.critic_optimizer_type,
            )

    def unwrap_models(self):
        if self.accelerator is not None:
            self.actor = self.accelerator.unwrap_model(self.actor)
            self.actor_target = self.accelerator.unwrap_model(self.actor_target)
            self.critic = self.accelerator.unwrap_model(self.critic)
            self.critic_target = self.accelerator.unwrap_model(self.critic_target)
            self.actor_optimizer = self.accelerator.unwrap_model(self.actor_optimizer)
            self.critic_optimizer = self.accelerator.unwrap_model(self.critic_optimizer)

    def saveCheckpoint(self, path):
        """Saves a checkpoint of agent properties and network weights to path.

        :param path: Location to save checkpoint at
        :type path: string
        """
        torch.save(
            {
                "actor_init_dict": self.actor.init_dict,
                "actor_state_dict": self.actor.state_dict(),
                "actor_target_init_dict": self.actor_target.init_dict,
                "actor_target_state_dict": self.actor_target.state_dict(),
                "critic_init_dict": self.critic.init_dict,
                "critic_state_dict": self.critic.state_dict(),
                "critic_target_init_dict": self.critic_target.init_dict,
                "critic_target_state_dict": self.critic_target.state_dict(),
                "actor_optimizer_state_dict": self.actor_optimizer.state_dict(),
                "critic_optimizer_state_dict": self.critic_optimizer.state_dict(),
                "net_config": self.net_config,
                "batch_size": self.batch_size,
                "lr": self.lr,
                "learn_step": self.learn_step,
                "gamma": self.gamma,
                "tau": self.tau,
                "mutation": self.mut,
                "index": self.index,
                "scores": self.scores,
                "fitness": self.fitness,
                "steps": self.steps,
            },
            path,
            pickle_module=dill,
        )

    def loadCheckpoint(self, path):
        """Loads saved agent properties and network weights from checkpoint.

        :param path: Location to load checkpoint from
        :type path: string
        """
        checkpoint = torch.load(path, pickle_module=dill)
        self.net_config = checkpoint["net_config"]
        if self.net_config is not None:
            self.arch = checkpoint["net_config"]["arch"]
            if self.arch == "mlp":
                self.actor = EvolvableMLP(**checkpoint["actor_init_dict"])
                self.actor_target = EvolvableMLP(**checkpoint["actor_target_init_dict"])
                self.critic = EvolvableMLP(**checkpoint["critic_init_dict"])
                self.critic_target = EvolvableMLP(
                    **checkpoint["critic_target_init_dict"]
                )
            elif self.arch == "cnn":
                self.actor = EvolvableCNN(**checkpoint["actor_init_dict"])
                self.actor_target = EvolvableCNN(**checkpoint["actor_target_init_dict"])
                self.critic = EvolvableCNN(**checkpoint["critic_init_dict"])
                self.critic_target = EvolvableCNN(
                    **checkpoint["critic_target_init_dict"]
                )
        else:
            self.actor = MakeEvolvable(**checkpoint["actor_init_dict"])
            self.actor_target = MakeEvolvable(**checkpoint["actor_target_init_dict"])
            self.critic = MakeEvolvable(**checkpoint["critic_init_dict"])
            self.critic_target = MakeEvolvable(**checkpoint["critic_target_init_dict"])
        self.lr = checkpoint["lr"]
        self.actor_optimizer = optim.Adam(self.actor.parameters(), lr=self.lr)
        self.critic_optimizer = optim.Adam(self.critic.parameters(), lr=self.lr)
        self.actor.load_state_dict(checkpoint["actor_state_dict"])
        self.actor_target.load_state_dict(checkpoint["actor_target_state_dict"])
        self.critic.load_state_dict(checkpoint["critic_state_dict"])
        self.critic_target.load_state_dict(checkpoint["critic_target_state_dict"])
        self.actor_optimizer.load_state_dict(checkpoint["actor_optimizer_state_dict"])
        self.critic_optimizer.load_state_dict(checkpoint["critic_optimizer_state_dict"])
        self.batch_size = checkpoint["batch_size"]
        self.learn_step = checkpoint["learn_step"]
        self.gamma = checkpoint["gamma"]
        self.tau = checkpoint["tau"]
        self.mut = checkpoint["mutation"]
        self.index = checkpoint["index"]
        self.scores = checkpoint["scores"]
        self.fitness = checkpoint["fitness"]
        self.steps = checkpoint["steps"]



import json
import pickle as pkl
import random
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
from tqdm.auto import tqdm

from agilerl.data.language_environment import Policy, interact_environment
from agilerl.data.rl_data import (
    ConstantTokenReward,
    DataPoint,
    Iterable_RL_Dataset,
    List_RL_Dataset,
    TokenReward,
)
from agilerl.utils.ilql_utils import convert_path
from agilerl.wordle.wordle_env import WordleEnvironment, WordleObservation
from agilerl.wordle.wordle_game import Vocabulary, WordleGame, WordleState
from agilerl.wordle.wordle_tokenizer import WordleTokenizer


class WordleListDataset(List_RL_Dataset):
    def __init__(
        self,
        items: List[Tuple[WordleObservation, Optional[Dict[str, Any]]]],
        max_len: Optional[int],
        token_reward: TokenReward,
    ) -> None:
        tokenizer = WordleTokenizer()
        super().__init__(tokenizer, token_reward, max_len)
        self.items = items

    def get_item(self, idx: int):
        return DataPoint.from_obs(
            self.items[idx][0], self.tokenizer, self.token_reward, self.items[idx][1]
        )

    def size(self):
        return len(self.items)

    @classmethod
    def from_file(
        cls,
        file_path: str,
        max_len: Optional[int],
        vocab: Optional[Vocabulary],
        token_reward: TokenReward,
    ):
        with open(file_path, "rb") as f:
            d = pkl.load(f)
        if vocab is None:
            vocab = Vocabulary.from_file(convert_path(d["vocab_path"]))
            if d["vocab_cache_path"] is not None:
                vocab.cache.load(convert_path(d["vocab_cache_path"]))
        wordle_items = [
            WordleObservation(
                WordleGame(
                    item["state"], vocab.update_vocab(item["state"]), item["actions"]
                )
            )
            for item in tqdm(d["state_actions"])
        ]
        meta = [
            {**item["meta"], "self": wordle_items[i]}
            if "meta" in item
            else {"self": wordle_items[i]}
            for i, item in enumerate(d["state_actions"])
        ]
        return WordleListDataset(list(zip(wordle_items, meta)), max_len, token_reward)


class WordleIterableDataset(Iterable_RL_Dataset):
    def __init__(
        self,
        policy: Policy,
        vocab: Vocabulary,
        max_len: Optional[int],
        token_reward: TokenReward,
    ) -> None:
        tokenizer = WordleTokenizer()
        super().__init__(tokenizer, token_reward, max_len)
        self.policy = policy
        self.env = WordleEnvironment(vocab)

    def sample_item(self):
        return DataPoint.from_obs(
            interact_environment(self.env, self.policy, None)[0],
            self.tokenizer,
            self.token_reward,
            None,
        )


class WordleHumanDataset(Iterable_RL_Dataset):
    def __init__(
        self,
        games: List[Tuple[str, List[str]]],
        transitions: Dict[str, Dict[str, List[str]]],
        use_true_word: bool,
        max_len: Optional[int],
        token_reward: TokenReward,
        game_indexes: Optional[List[int]],
        top_p: Optional[float],
    ) -> None:
        tokenizer = WordleTokenizer()
        super().__init__(tokenizer, token_reward, max_len)
        self.games = games
        if game_indexes is not None:
            self.games = [self.games[idx] for idx in game_indexes]
        if top_p is not None:
            lens = [len(game) for _, game in self.games]
            self.games = [
                self.games[idx] for idx in np.argsort(lens)[: int(len(lens) * top_p)]
            ]
        self.transitions = transitions
        self.use_true_word = use_true_word

    def sample_item(self):
        true_word, game = random.choice(self.games)
        if self.use_true_word:
            while True:
                actions = []
                for transition in game:
                    if (
                        transition not in self.transitions[true_word]
                        or len(self.transitions[true_word][transition]) == 0
                    ):
                        break
                    actions.append(
                        random.choice(self.transitions[true_word][transition])
                    )
                if len(actions) == len(game):
                    break
                else:
                    true_word, game = random.choice(self.games)
        else:
            word_choices = list(self.transitions.keys())
            while True:
                true_word = random.choice(word_choices)
                actions = []
                for transition in game:
                    if (
                        transition not in self.transitions[true_word]
                        or len(self.transitions[true_word][transition]) == 0
                    ):
                        break
                    actions.append(
                        random.choice(self.transitions[true_word][transition])
                    )
                if len(actions) == len(game):
                    break
                else:
                    true_word, game = random.choice(self.games)
        state = WordleState.initial_state()
        for action in actions:
            state = state.transition_state(action, true_word)
        vocab = Vocabulary([true_word], state, cache=None, fill_cache=False)
        obs = WordleObservation(WordleGame(state, vocab, actions))
        return DataPoint.from_obs(obs, self.tokenizer, self.token_reward, {"obs": obs})

    @classmethod
    def from_file(
        cls,
        file_path: str,
        use_true_word: bool = False,
        max_len: Optional[int] = None,
        token_reward: Optional[TokenReward] = None,
        game_indexes: Optional[List[int]] = None,
        top_p: Optional[float] = None,
    ):
        if token_reward is None:
            token_reward = ConstantTokenReward(0.0)
        with open(file_path) as f:
            d = json.load(f)
        return WordleHumanDataset(
            d["games"],
            d["transitions"],
            use_true_word,
            max_len,
            token_reward,
            game_indexes,
            top_p,
        )
